"""
    This file shows how machine learning can be wrapped into trading decision very easily.

    A model is trained in the init() function, then prediction is run in the price event. This is simulated to run
    once a day
"""

import numpy as np
from blankly import trunc
from blankly import Strategy, StrategyState, Interface
from blankly import CoinbasePro
from blankly.indicators import rsi, sma
# You may need to "pip install scikit-learn" if you do not have this installed
from sklearn.neural_network import MLPClassifier
from sklearn.datasets import make_classification
from sklearn.preprocessing import MinMaxScaler
from sklearn.model_selection import train_test_split


def init(symbol, state: StrategyState):
    interface: Interface = state.interface
    resolution = state.resolution
    variables = state.variables
    x, y = make_classification(n_samples=500, n_features=3, n_informative=3, n_redundant=0, random_state=1)
    x_train, x_test, y_train, y_test = train_test_split(x, y, stratify=y, random_state=1)
    # initialize the historical data
    variables['model'] = MLPClassifier(random_state=1, max_iter=10000).fit(x_train, y_train)
    variables['history'] = interface.history(symbol, 300, resolution, return_as='list')['close']
    variables['has_bought'] = False


def price_event(price, symbol, state: StrategyState):
    interface: Interface = state.interface
    variables = state.variables

    variables['history'].append(price)
    model = variables['model']
    scaler = MinMaxScaler()
    rsi_values = rsi(variables['history'], period=14).reshape(-1, 1)
    rsi_value = scaler.fit_transform(rsi_values)[-1]

    ma_values = sma(variables['history'], period=50).reshape(-1, 1)
    ma_value = scaler.fit_transform(ma_values)[-1]

    ma100_values = sma(variables['history'], period=100).reshape(-1, 1)
    ma100_value = scaler.fit_transform(ma100_values)[-1]
    value = np.array([rsi_value, ma_value, ma100_value]).reshape(1, 3)
    prediction = model.predict_proba(value)[0][1]

    # comparing prev diff with current diff will show a cross
    if prediction > 0.4 and not variables['has_bought']:
        interface.market_order(symbol, 'buy', trunc(interface.cash/price, 8))
        variables['has_bought'] = True
    elif prediction <= 0.4 and variables['has_bought']:
        # truncate is required due to float precision
        interface.market_order(symbol, 'sell', interface.account[state.base_asset]['available'])
        variables['has_bought'] = False


if __name__ == "__main__":
    coinbase = CoinbasePro()
    s = Strategy(coinbase)
    # creating an init allows us to run the same function for
    # different tickers and resolutions
    s.add_price_event(price_event, 'BTC-USD', resolution='1d', init=init)
    # s.add_price_event(price_event, 'AAPL', resolution='1d', init=init)
    history = s.backtest(to='1y', initial_values={'USD': 100})
    print(history)



from blankly.exchanges.interfaces.binance_futures.binance_futures import BinanceFutures
from blankly.exchanges.interfaces.ftx_futures.ftx_futures import FTXFutures
from blankly.frameworks.strategy import FuturesStrategy
from blankly.frameworks.strategy import FuturesStrategyState

from blankly.enums import MarginType, HedgeMode


"""
    Binance Futures Exchange object.
    Copyright (C) 2021 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from binance.client import Client

from blankly.exchanges.auth.auth_constructor import AuthConstructor
from blankly.exchanges.futures.futures_exchange import FuturesExchange
from blankly.exchanges.interfaces.binance_futures.binance_futures_interface import BinanceFuturesInterface
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface


class BinanceFutures(FuturesExchange):

    def __init__(self,
                 portfolio_name=None,
                 keys_path="keys.json",
                 preferences_path=None):
        super().__init__("binance_futures", portfolio_name, preferences_path)

        # Load auth from keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'binance',
                               ['API_KEY', 'API_SECRET', 'sandbox'])

        self.__calls = Client(api_key=auth.keys['API_KEY'],
                              api_secret=auth.keys['API_SECRET'],
                              testnet=auth.keys['sandbox'])

        self.__interface = BinanceFuturesInterface(self.exchange_type,
                                                   self.calls)

        self.initialize()

    @property
    def calls(self):
        return self.__calls

    @property
    def interface(self) -> FuturesExchangeInterface:
        return self.__interface



"""
    Comparison utilities for tests
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


def is_sub_dict(dict1: dict, dict2: dict) -> bool:
    for key in dict1:
        if dict1[key] != dict2.get(key):
            return False

    return True


def validate_response(needed, resp):
    # verify resp is correct
    for key_type in needed:
        key = key_type[0]
        val_type = key_type[1]

        assert key in resp, f'{key} not in response'
        assert isinstance(resp[key], val_type), f'{key} has wrong value type'



"""
    Paper trading utilities for generating accurate paper trading responses & behaviors.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import secrets


def generate_coinbase_pro_id():
    # Create coinbase pro-like id
    coinbase_pro_id = secrets.token_hex(nbytes=16)
    coinbase_pro_id = coinbase_pro_id[:8] + '-' + coinbase_pro_id[8:]
    coinbase_pro_id = coinbase_pro_id[:13] + '-' + coinbase_pro_id[13:]
    coinbase_pro_id = coinbase_pro_id[:18] + '-' + coinbase_pro_id[18:]
    coinbase_pro_id = coinbase_pro_id[:23] + '-' + coinbase_pro_id[23:]

    return coinbase_pro_id



"""
    FTX definition & setup
    Copyright (C) 2021 Blankly Finance

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from blankly.exchanges.exchange import Exchange
from blankly.exchanges.interfaces.ftx.ftx_api import FTXAPI
from blankly.exchanges.auth.auth_constructor import AuthConstructor
from blankly.exchanges.interfaces.paper_trade.paper_trade_interface import PaperTradeInterface
from blankly.exchanges.interfaces.ftx.ftx_interface import FTXInterface
from blankly.utils import utils


class FTX(Exchange):
    def __init__(self, portfolio_name=None, keys_path="keys.json", settings_path=None):
        Exchange.__init__(self, "ftx", portfolio_name, settings_path)

        # Load the auth from the keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'ftx', ['API_KEY', 'API_SECRET', 'sandbox'])

        keys = auth.keys
        sandbox = super().evaluate_sandbox(auth)

        calls = FTXAPI(keys['API_KEY'], keys['API_SECRET'],
                       tld=self.preferences["settings"]["ftx"]["ftx_tld"])

        # Always finish the method with this function
        super().construct_interface_and_cache(calls)

        # FTX is unique because we can continue by wrapping the interface in paper trade
        if sandbox:
            utils.info_print('The sandbox setting is enabled for this key. FTX has been created as a '
                             'paper trading instance.')
            self.interface = PaperTradeInterface(FTXInterface('ftx', calls))

    """
    Builds information about the asset on this exchange by making particular API calls
    """

    def get_asset_state(self, symbol):
        """
        This determines the internal properties of the exchange block.
        Should be implemented per-class because it requires different types of interaction with each exchange.
        """
        # TODO Populate this with useful information
        return self.interface.get_account(symbol)

    def get_exchange_state(self):
        """
        Exchange state is the external properties for the exchange block
        """
        # TODO Populate this with useful information
        return self.interface.get_fees()

    def get_direct_calls(self) -> FTXAPI:
        return self.calls



import datetime
import time
import typing
import warnings

import pandas as pd

import blankly
from blankly.exchanges.exchange import Exchange
from blankly.exchanges.interfaces.paper_trade.backtest_controller import BackTestController
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.exchanges.strategy_logger import StrategyLogger
from blankly.frameworks.strategy import StrategyBase
from blankly.utils.time_builder import time_interval_to_seconds
from blankly.utils.utils import info_print


class Strategy(StrategyBase):
    _exchange: Exchange

    def __init__(self, exchange: Exchange):
        super().__init__(exchange, StrategyLogger(exchange.get_interface(), strategy=self))
        self._paper_trade_exchange = blankly.PaperTrade(self._exchange)

    def backtest(self,
                 to: str = None,
                 initial_values: dict = None,
                 start_date: typing.Union[str, float, int] = None,
                 end_date: typing.Union[str, float, int] = None,
                 save: bool = False,
                 settings_path: str = None,
                 callbacks: list = None,
                 **kwargs
                 ) -> BacktestResult:
        """
        Turn this strategy into a backtest.

        Args:
            ** We expect either an initial_value (in USD) or a dictionary of initial values, we also expect
            either `to` to be set or `start_date` and `end_date` **

            to (str): Declare an amount of time before now to backtest from: ex: '5y' or '10h'
            initial_values (dict): Dictionary of initial value sizes (i.e { 'BTC': 3, 'USD': 5650}).
                Using this will override the values that are currently on your exchange.
            start_date (str): Override argument "to" by specifying a start date such as "03/06/2018". This can also
                be an epoch time as a float or int.
            end_date (str): End the backtest at a date such as "03/06/2018". This can also be an epoch type as a float
                or int
            save (bool): Save the price data references to the data required for the backtest as well as
                overriden settings.
            settings_path (str): Path to the backtest.json file.
            callbacks (list of callables): Custom functions that will be run at the end of the backtest

            Keyword Arguments:
                **Use these to override parameters in the backtest.json file**

                use_price: str = 'close',
                    Set which price column to use.

                smooth_prices: bool = False,
                    Create linear connections between downloaded prices

                GUI_output: bool = True,
                    Enable/disable GUI webpage display after backtest

                show_tickers_with_zero_delta: bool = False,
                    Exclude tickers that have no change to account value in the GUI

                save_initial_account_value: bool = True,
                    Put an extra frame which contains the initial account values before any trade

                show_progress_during_backtest: bool = True,
                    Show a progress bar as the backtest runs

                cache_location: str = './price_caches'
                    Set a location for the price cache csv's to be written to

                continuous_caching: bool
                    Utilize the advanced price caching system built into the backtest. Automatically aggregate and prune
                    downloaded data.

                resample_account_value_for_metrics: str or bool = '1d' or False
                    Because backtest data can be input at a variety of resolutions, account value often needs to be
                        recalculated at consistent intervals for use in metrics & indicators.
                        This setting allows the specification of that consistent interval.
                        The value can be set to `False` to skip any recalculation.

                quote_account_value_in: str = 'USD'
                    Manually set what valuation should be used when calculating account value.
                        Multiple types of quote currency (ex: USD and EUR) are not supported because
                        there is no datasource for quoting pairs such as EUR-USD until forex integration.

                ignore_user_exceptions: bool = True
                    Set this to True to handle user exceptions identically to how they're handled by strategy calls.
                        False means that the backtest will immediately stop & attempt to generate a report if something
                        in the user calls goes wrong. True will replicate strategy errors.

                risk_free_return_rate: float = 0.0
                    Set this to be the theoretical rate of return with no risk
        """

        start = None
        end = None

        # Even if they specified start/end unevenly it will be overwritten with any to argument
        if to is not None:
            start = time.time() - time_interval_to_seconds(to)
            end = time.time()

        if start_date is not None:
            if isinstance(end_date, (int, float)):
                start = start_date
            else:
                start_date = pd.to_datetime(start_date)
                epoch = datetime.datetime.utcfromtimestamp(0)
                start = (start_date - epoch).total_seconds()

        if end_date is not None:
            if isinstance(end_date, (int, float)):
                end = end_date
            else:
                end_date = pd.to_datetime(end_date)
                epoch = datetime.datetime.utcfromtimestamp(0)
                end = (end_date - epoch).total_seconds()

        # If start/ends are specified unevenly
        if (start_date is None and end_date is not None) or (start_date is not None and end_date is None):
            raise ValueError("Both start and end dates must be set or use the 'to' argument.")

        self.interface = self._paper_trade_exchange.get_interface()
        self.backtesting_controller = BackTestController(self._paper_trade_exchange,
                                                         backtest_settings_path=settings_path,
                                                         callbacks=callbacks
                                                         )

        # Write any kwargs as settings to the settings - save if enabled.
        for k, v in kwargs.items():
            self.backtesting_controller.write_setting(k, v, save)

        if initial_values is not None:
            self.backtesting_controller.write_initial_price_values(initial_values)

        # Write each scheduling pair as a price event - save if enabled.
        if start is not None and end is not None:
            for i in self._scheduling_pair:
                # None means live which is orderbook, which we skip anyway
                if i[1] is not None:
                    self.backtesting_controller.add_prices(i[0], start, end, i[1], save=save)
        else:
            info_print("User-specified start and end time not given. Defaulting to using only cached data.")

        if self._using_orderbook:
            warning_string = "Artificial orderbook generation is not yet supported for backtesting - " \
                             "skipping orderbook callbacks."
            warnings.warn(warning_string)

        # Append each of the events the class defines into the backtest
        for i in self._schedulers:
            kwargs = i.get_kwargs()
            self.backtesting_controller.append_backtest_price_event(callback=kwargs['callback'],
                                                                    asset_id=kwargs['symbol'],
                                                                    time_interval=i.get_interval(),
                                                                    state_object=kwargs['state_object'],
                                                                    ohlc=kwargs['ohlc'],
                                                                    init=kwargs['init'],
                                                                    teardown=kwargs['teardown']
                                                                    )

        # Run the backtest & return results
        results = self.backtesting_controller.run()

        blankly.reporter.export_backtest_result(results)

        # Clean up
        self.interface = self._interface_cache
        return results

    def time(self) -> float:
        if self.backtesting_controller is not None and self.backtesting_controller.time is not None:
            return self.backtesting_controller.time
        else:
            return time.time()



"""
    Class to manage a stop order lifecycle - a way to determine individual information about each buy and sell
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from blankly.exchanges.orders.order import Order


class StopLimit(Order):
    """
    (Buying or selling (string), amount in currency (BTC/XLM), ticker object (so we can get time and value),
    limit if there is one)
    """

    def __init__(self, order, response, interface):
        """
        Coinbase Response:
        {
            'id': '3a98a5c6-05a0-4e46-b8e4-3f27358fe27d',
            'price': '29500',
            'size': '0.01',
            'product_id': 'BTC-USD',
            'side': 'buy',
            'stp': 'dc',
            'type': 'limit',
            'time_in_force':
            'GTC',
            'post_only': False,
            'created_at': '2021-05-28T19:24:52.010449Z',
            'stop': 'loss',
            'stop_price': '30000',
            'fill_fees': '0',
            'filled_size': '0',
            'executed_value': '0',
            'status': 'pending',
            'settled': False
        }
        binance Response:
        TODO add this

        Guaranteed:
        needed = [
                ["symbol", str],
                ["id", str],  <-- Handled by order
                ["created_at", float],  <-- Handled by order
                ["stop_price", float],
                ["limit_price", float],
                ["stop", str],
                ["size", float],
                ["status", str],  <-- Handled by order
                ["time_in_force", str],  <-- Handled by order
                ["type", str],  <-- Handled by order
                ["side", str]  <-- Handled by order
        ],
        """
        self.Interface = interface
        self.__exchange = self.Interface.get_exchange_type()
        self.__order = order
        self.__response = response
        self.__homogenized_result = None
        super().__init__(self.__response, self.__order, self.Interface)

    def get_stop_price(self) -> float:
        """
        Get the trigger price of the stop order
        """
        return self.__response["stop_price"]

    def get_limit_price(self) -> float:
        """
        Get the price that the limit is set at after triggering the stop
        """
        return self.__response["limit_price"]

    def get_stop_type(self) -> str:
        """
        Get if the stop type is "loss" or "entry"
        """
        return self.__response["stop"]

    def get_size(self) -> float:
        """
        Get the amount of base currency set in the triggered stop limit order
        """
        return self.__response["size"]

    def get_time_in_force(self) -> str:
        """
        Get the exchange's set time_in_force value.
        """
        return self.__response["time_in_force"]

    def __str__(self):
        return_string = super().__str__()

        return_string = self.add_new_line(return_string, "Stop Order Parameters: ")

        return_string = self.add_new_line(return_string, "Time In Force: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_time_in_force())

        return_string = self.add_new_line(return_string, "Stop Price: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_stop_price())

        return_string = self.add_new_line(return_string, "Stop Limit Price: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_limit_price())

        return_string = self.add_new_line(return_string, "Stop Type: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_stop_type())

        return_string = self.add_new_line(return_string, "Get Size: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_size())

        return return_string



"""
    Alpaca websocket definitions
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import collections
import ssl
import threading
import time
import traceback

import msgpack
from websocket import create_connection

import blankly
from blankly.utils.utils import info_print
from blankly.exchanges.abc_exchange_websocket import ABCExchangeWebsocket
from blankly.exchanges.auth.utils import load_auth
from blankly.exchanges.interfaces.alpaca.alpaca_websocket_utils import parse_alpaca_timestamp, switch_type


def subscribe(ws, channel, symbol):
    ws.send(
        msgpack.packb({
            'action': 'subscribe',
            channel: tuple([symbol]),
        })
    )


def unsubscribe(ws, channel):
    ws.send(
        msgpack.packb({
            'action': 'unsubscribe',
            channel: (),
        })
    )


def create_ticker_connection(symbol, url, channel):
    ws = create_connection(url, sslopt={"cert_reqs": ssl.CERT_NONE}, header={'Content-Type': 'application/msgpack'})
    _, auth = load_auth('alpaca')

    ws.send(msgpack.packb({
        'action': 'auth',
        'key': auth['API_KEY'],
        'secret': auth['API_SECRET']
    }))
    # Check if the response is valid
    response = msgpack.unpackb(ws.recv())
    if response[0]['msg'] != 'connected':
        info_print("Connection failed.")

    subscribe(ws, channel, symbol)
    response = msgpack.unpackb(ws.recv())
    if response[0]['msg'] != 'authenticated':
        if response[0]['msg'] == 'connection limit exceeded':
            # Unsubscribe and resubscribe
            unsubscribe(ws, channel)
            print(msgpack.unpackb(ws.recv()))
            subscribe(ws, channel, symbol)
            print(msgpack.unpackb(ws.recv()))
        else:
            info_print("Authentication failed.")

    return ws


# List of valid subscriptions
# 'trades': tuple(['AAPL']),
# 'quotes': tuple(['AAPL']),
# 'bars': tuple(['AAPL']),
# 'dailyBars': tuple(['AAPL']),
# 'statuses': tuple(['AAPL']),
# 'lulds': tuple(['AAPL'])


class Tickers(ABCExchangeWebsocket):
    def __init__(self, symbol, stream, log=None,
                 pre_event_callback=None, initially_stopped=False,
                 WEBSOCKET_URL="wss://stream.data.alpaca.markets/v2/iex/"):
        """
        Create and initialize the ticker
        Args:
            symbol: Currency to initialize on such as "BTC-USD"
            stream: Valid exchange stream to attach to, these are:
                trades, quotes, bars, dailyBars, statuses, lulds
            log: Fill this with a path to a log file that should be created
            WEBSOCKET_URL: Default websocket URL feed.
        """
        self.__symbol = symbol
        self.__stream = stream
        self.__logging_callback, self.__interface_callback, log_message = switch_type(stream)

        # Initialize log file
        if log is not None:
            self.__log = True
            self.__filePath = log
            try:
                self.__file = open(log, 'x+')
                self.__file.write(log_message)
            except FileExistsError:
                self.__file = open(log, 'a')
        else:
            self.__log = False

        self.URL = WEBSOCKET_URL
        self.ws = None
        self.__response = None
        self.__subscription_response = None
        self.__most_recent_tick = None
        self.__most_recent_time = None
        self.__callbacks = []
        self.__pre_event_callback = pre_event_callback

        # Reload preferences
        self.__preferences = blankly.utils.load_user_preferences()
        buffer_size = self.__preferences["settings"]["websocket_buffer_size"]
        self.__ticker_feed = collections.deque(maxlen=buffer_size)
        self.__time_feed = collections.deque(maxlen=buffer_size)

        # Start the websocket
        if not initially_stopped:
            self.start_websocket()

    def start_websocket(self):
        """
        Restart websocket if it was asked to stop.
        """
        if self.ws is None:
            self.ws = create_ticker_connection(self.__symbol, self.URL, self.__stream)
            self.__response = self.ws.recv()
            thread = threading.Thread(target=self.read_websocket)
            thread.start()
        else:
            if self.ws.connected:
                print("Already running...")
                pass
            else:
                # Use recursion to restart, continue appending to time feed and ticker feed
                self.ws = None
                self.start_websocket()

    def read_websocket(self):
        counter = 0
        # TODO port this to "WebSocketApp" found in the websockets documentation
        while self.ws.connected:
            # In case the user closes while its reading from the websocket, this will let it expire
            persist_connected = self.ws.connected
            try:
                received = self.ws.recv()
                received = msgpack.unpackb(received)[0]  # type: dict
                # Modify time to use epoch

                if 't' in received:
                    received['t'] = parse_alpaca_timestamp(received['t'])
                    recent_time = received['t']
                elif 'code' in received:
                    continue
                else:
                    recent_time = time.time()

                if self.__log:
                    if counter % 100 == 0:
                        self.__file.close()
                        self.__file = open(self.__filePath, 'a')
                    line = self.__logging_callback(received)
                    self.__file.write(line)

                # Manage price events and fire for each manager attached
                interface_message = self.__interface_callback(received)

                self.__most_recent_time = recent_time
                self.__time_feed.append(self.__most_recent_time)
                self.__most_recent_tick = interface_message
                self.__ticker_feed.append(interface_message)

                try:
                    for i in self.__callbacks:
                        i(interface_message)
                except Exception:
                    traceback.print_exc()

                counter += 1
            except Exception:
                traceback.print_exc()
                if persist_connected:
                    pass
                else:
                    print("Error reading ticker websocket for " + self.__symbol + " on " +
                          self.__stream + ": attempting to re-initialize")
                    # Give a delay so this doesn't eat up from the main thread if it takes many tries to initialize
                    time.sleep(2)
                    self.ws.close()
                    self.ws = create_ticker_connection(self.__symbol, self.URL, self.__stream)
                    # Update response
                    self.__response = self.ws.recv()

    """ Required in manager """

    def is_websocket_open(self):
        return self.ws.connected

    def get_currency_id(self):
        return self.__symbol

    """ Required in manager """

    def append_callback(self, obj):
        self.__callbacks.append(obj)

    """ Define a variable each time so there is no array manipulation """
    """ Required in manager """

    def get_most_recent_tick(self):
        return self.__most_recent_tick

    """ Required in manager """

    def get_most_recent_time(self):
        return self.__most_recent_time

    """ Required in manager """

    def get_time_feed(self):
        return list(self.__time_feed)

    """ Parallel with time feed """
    """ Required in manager """

    def get_feed(self):
        return list(self.__ticker_feed)

    """ Required in manager """

    def get_response(self):
        return self.__response

    """ Required in manager """

    def close_websocket(self):
        if self.ws.connected:
            self.ws.close()
        else:
            print("Websocket for " + self.__symbol + ' on channel ' + self.__stream + " is already closed")

    """ Required in manager """

    def restart_ticker(self):
        self.start_websocket()



"""
    Statistics functions
    Copyright (C) 2021 Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any

import pandas as pd
import tulipy as ti

from blankly.indicators.utils import check_series, convert_to_numpy


def stddev_period(data, period=14, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    stddev = ti.stddev(data, period)
    return pd.Series(stddev) if use_series else stddev


def var_period(data, period=14, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    var = ti.var(data, period)
    return pd.Series(var) if use_series else var


def stderr_period(data, period=14, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    stderr = ti.stderr(data, period)
    return pd.Series(stderr) if use_series else stderr


def min_period(data, period, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    minimum = ti.min(data, period)
    return pd.Series(minimum) if use_series else minimum


def max_period(data, period, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    maximum = ti.max(data, period)
    return pd.Series(maximum) if use_series else maximum


def sum_period(data, period, use_series=False) -> Any:
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    maximum = ti.sum(data, period)
    return pd.Series(maximum) if use_series else maximum



import blankly


def price_event(price, symbol, state: blankly.StrategyState):
    """ This function will give an updated price every 15 seconds from our definition below """
    state.variables['history'].append(price)
    rsi = blankly.indicators.rsi(state.variables['history'])
    if rsi[-1] < 30 and not state.variables['owns_position']:
        # Dollar cost average buy
        buy = blankly.trunc(state.interface.cash/price, 2)
        state.interface.market_order(symbol, side='buy', size=buy)
        state.variables['owns_position'] = True
    elif rsi[-1] > 70 and state.variables['owns_position']:
        # Dollar cost average sell
        curr_value = state.interface.account[state.base_asset].available
        state.interface.market_order(symbol, side='sell', size=curr_value)
        state.variables['owns_position'] = False


def init(symbol, state: blankly.StrategyState):
    # Download price data to give context to the algo
    state.variables['history'] = state.interface.history(symbol, to=150, return_as='deque',
                                                         resolution=state.resolution)['close']
    state.variables['owns_position'] = False


if __name__ == "__main__":
    # Authenticate coinbase pro strategy
    exchange = blankly.CoinbasePro()

    # Use our strategy helper on coinbase pro
    strategy = blankly.Strategy(exchange)

    # Run the price event function every time we check for a new price - by default that is 15 seconds
    strategy.add_price_event(price_event, symbol='BTC-USD', resolution='1d', init=init)

    # Start the strategy. This will begin each of the price event ticks
    # strategy.start()
    # Or backtest using this
    results = strategy.backtest(to='1y', initial_values={'USD': 10000})
    print(results)



from blankly import Screener, Alpaca, ScreenerState
from blankly.indicators import rsi

tickers = ['AAPL', 'GME', 'MSFT']  # any stocks that you may want


# This function is our evaluator and runs per stock
def is_stock_buy(symbol, state: ScreenerState):
    # This runs per stock
    prices = state.interface.history(symbol, 40, resolution='1d',
                                     return_as='list')  # get past 40 data points
    price = state.interface.get_price(symbol)
    rsi_values = rsi(prices['close'], 14)
    return {'is_oversold': bool(rsi_values[-1] < 30), 'price': price, 'symbol': symbol}

def formatter(results, state: ScreenerState):
    # results is a dictionary on a per-symbol basis
    result_string = 'These are all the stocks that are currently oversold: \n'
    for symbol in results:
        if results[symbol]['is_oversold']:
            result_string += '{} is currently oversold at a price of {}\n\n'.format(symbol, results[symbol]['price'])
    return result_string


if __name__ == "__main__":
    alpaca = Alpaca()  # initialize our interface
    screener = Screener(alpaca, is_stock_buy, symbols=tickers, formatter=formatter)  # find oversold

    print(screener.formatted_results)



"""
    Backtesting controller for paper trading interface
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import os
import time
import traceback
import typing
from datetime import datetime as dt
import copy

import numpy as np
import pandas as pd
import requests
from bokeh.layouts import column as bokeh_columns
from bokeh.models import HoverTool
from bokeh.palettes import Category10_10
from bokeh.plotting import ColumnDataSource, figure, show

import blankly.exchanges.interfaces.paper_trade.metrics as metrics
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.exchanges.interfaces.paper_trade.paper_trade import PaperTrade
from blankly.exchanges.interfaces.paper_trade.paper_trade_interface import PaperTradeInterface
from blankly.utils.time_builder import time_interval_to_seconds
from blankly.utils.utils import load_backtest_preferences, update_progress, write_backtest_preferences, \
    get_base_asset, get_quote_asset, info_print
from blankly.exchanges.interfaces.paper_trade.backtest.format_platform_result import \
    format_platform_result


def to_string_key(separated_list):
    output = ""
    for i in range(len(separated_list) - 1):
        output += str(separated_list[i])
        output += ","
    output += str(separated_list[-1:][0])
    return output


def split(base_range, local_segments) -> typing.Tuple[list, list]:
    """
    Find the negative given from a range and a set of other ranges

    Args:
        base_range: Backing array containing a range such as [1, 10]
        local_segments: An array of segments that we have such as [-5, 5], [6,7]
    Returns:
        The output of the example inputs above would be [[8, 10]]
    """

    # If we don't have any local segments there is no need for any of this, just download the whole set
    used_ranges = []
    positive_ranges = []  # These are the ranges that we have downloaded
    negative_ranges = []  # These are the ranges that we need

    def intersection(as_, ae, bs, be):
        """
        Find the intersection of two ranges
        Implementation from https://scicomp.stackexchange.com/a/26260

        Args:
            as_: Low of first range
            ae: High of first range
            bs: Low of second range
            be: High of second range
        """
        if bs > ae or as_ > be:
            return None
        else:
            os_ = max(as_, bs)
            oe = min(ae, be)
            return [os_, oe]
    for i in local_segments:
        intersection_range = intersection(base_range[0], base_range[1], i[0], i[1])
        if intersection_range is None:
            continue
        # We now know every single downloaded range that we need to keep
        used_ranges.append(i)

        # Now these are all intersections that we will filter to find the negative ranges to download
        positive_ranges.append(intersection_range)

    # Create a list of indexes that need to be pushed into a single list, arrays like  [1, 5] and [3, 7] would need to
    #  become [1, 7]

    # First we have to sort the lists to have the same starting value
    positive_ranges = sorted(positive_ranges, key=lambda x: x[0])

    aggregate_indexes = []
    for i in range(len(positive_ranges) - 1):
        intersection_range = intersection(positive_ranges[i][0], positive_ranges[i][1],
                                          positive_ranges[i + 1][0],     positive_ranges[i + 1][1])
        if intersection_range is not None:
            aggregate_indexes.append([i, i + 1])

    # Now aggregate from the last result
    for i in aggregate_indexes:
        # Pull the lower bound right out of the lower index from the aggregate_indexes approach
        lower_bound = positive_ranges[i[0]][0]

        positive_ranges[i[0]] = None

        # Push that same lower bound into the array that was above it
        positive_ranges[i[1]][0] = lower_bound

    # Filter out the None's that were added
    positive_ranges = [x for x in positive_ranges if not (x is None)]

    # Now just try to find the gaps in the positive ranges and those are our negative ranges
    for i in range(len(positive_ranges) - 1):
        negative_ranges.append([positive_ranges[i][1], positive_ranges[i+1][0]])

    try:
        # Now we just have to make sure to check the bounds
        # If the very first value we have is greater than what we requested
        if positive_ranges[0][0] > base_range[0]:
            # We then need the base range up to our first positive value
            negative_ranges.append([base_range[0], positive_ranges[0][0]])

        # If the last positive range is less than our requested final
        if positive_ranges[-1][1] < base_range[1]:
            # The actual value is from the last positive range value to our requested final value
            negative_ranges.append([positive_ranges[-1][1], base_range[1]])
    except IndexError:
        return [], [base_range]

    return used_ranges, negative_ranges


class BackTestController:
    def __init__(self, paper_trade_exchange: PaperTrade, backtest_settings_path: str = None, callbacks: list = None):
        self.preferences = load_backtest_preferences(backtest_settings_path)
        self.backtest_settings_path = backtest_settings_path
        if not paper_trade_exchange.get_type() == "paper_trade":
            raise ValueError("Backtest controller was not constructed with a paper trade exchange object.")
        self.interface: PaperTradeInterface = paper_trade_exchange.get_interface()

        self.price_events = []

        if callbacks is None:
            callbacks = []
        self.callbacks = callbacks

        self.current_time = None
        self.initial_time = None

        self.prices = []  # [epoch, "BTC-USD", price]

        self.price_dictionary = {}

        self.pd_prices = None

        # User added times
        self.__user_added_times = []

        self.use_price = None

        self.queue_backtest_write = False

        self.quote_currency = None

        # Create a global generator because a second yield function gets really nasty
        self.__color_generator = Category10_10.__iter__()

        self.initial_account = None

        self.__exchange_type = self.interface.get_exchange_type()

        # Create our own traded assets' dictionary because we customize it a bit
        self.__traded_assets = []

        # Because the times are run in order we can use this variable to optimize account value searching
        self.__current_search_index = 0

        # Export a time for use in other classes
        self.time = None

        # Set these when prices are added to find the first price and the last price
        self.user_start = None
        self.user_stop = None

        self.min_resolution = None

    def sync_prices(self) -> dict:
        """
        Parse the local file cache for the requested data, if it doesn't exist, request it from the exchange

        args:
            items: list of lists organized as ['symbol', 'start_time', 'end_time', 'resolution']

        returns:
            dictionary with keys for each 'symbol'
        """
        
        cache_folder = self.preferences['settings']["cache_location"]
        # Make sure the cache folder exists and read files
        try:
            files = os.listdir(cache_folder)
        except FileNotFoundError:
            files = []
            os.mkdir(cache_folder)

        available_files = []
        for i in range(len(files)):
            # example file name: 'BTC-USD.1622400000.1622510793.60.csv'
            # Remove the .csv from each of the files: BTC-USD.1622400000.1622510793.60
            identifier = files[i][:-4].split(",")
            # Cast to float first before
            try:
                identifier[1] = int(float(identifier[1]))
                identifier[2] = int(float(identifier[2]))
                identifier[3] = int(float(identifier[3]))
                available_files.append(identifier)
            except IndexError:
                raise IndexError(f"Please remove file {files[i]} in your price caches or "
                                 f"match the blankly cache format.")

        # This is only the downloaded data
        local_history_blocks = {}
        for i in available_files:
            # Sort these into resolutions
            asset = i[0]

            # Make sure to add the asset as a key
            if asset not in list(local_history_blocks.keys()):
                local_history_blocks[asset] = {}

            # Add each resolution to its own internal list
            if i[3] not in list(local_history_blocks[asset].keys()):
                local_history_blocks[asset][i[3]] = [i]
            else:
                local_history_blocks[asset][i[3]].append(i)

        # def segment(conjoined_ranges: list) -> list:
        #     """
        #     Split a list with sets of even intervals into multiple lists. The list must be single increments in all
        #     places except for the splits
        #     """
        #     ranges = []
        #     segments = 0
        #     for i in range(len(conjoined_ranges)-1):
        #         i += 1
        #         if conjoined_ranges[i] - conjoined_ranges[i-1] != 1:
        #             segmented_list = conjoined_ranges[segments:i]
        #
        #             ranges.append((segmented_list[0], segmented_list[-1]))
        #             segments = i + 1
        #
        #     return ranges

        # This is the data the user has requested: [asset_id, start_time, end_time, resolution]
        items = self.__user_added_times

        final_prices = {}
        for i in range(len(items)):
            if items[i] is None:
                continue
            asset = items[i][0]
            resolution = items[i][3]
            start_time = items[i][1]
            end_time = items[i][2] - resolution

            if end_time < start_time:
                raise RuntimeError("Must specify a longer timeframe to run the backtest.")

            download_ranges = []

            # Attempt to find the same symbol/asset possibilities in the backtest blocks
            try:
                # Make sure to copy it because if you don't you delete this for any similar resolution
                # If you don't copy it fails if you use two price events at the same resolution
                download_ranges = copy.deepcopy(local_history_blocks[asset][resolution])
            except KeyError:
                pass

            for j in range(len(download_ranges)):
                download_ranges[j] = download_ranges[j][1:3]

            used_ranges, negative_ranges = split([start_time, end_time], download_ranges)

            # for j in download_ranges:
            #     download_start_time = j[1]
            #     download_end_time = j[2] - resolution
            #
            #     contains_start = False
            #     contains_end = False
            #     if download_start_time <= start_time <= download_end_time:
            #         contains_start = True
            #     elif download_start_time <= end_time <= download_end_time:
            #         contains_end = True
            #
            #     if contains_start or contains_end:
            #         # Read in the whole thing if it has any sort of relevance on the data
            #         relevant_df = pd.read_csv(os.path.join(cache_folder,
            #                                                to_string_key(j) + ".csv"))
            #
            #         # If it only contains the start then it must be offset too high or is too long
            #         if contains_start and not contains_end:
            #             relevant_df = relevant_df[relevant_df['time'] >= start_time]
            #
            #             # Set the end time to match the first time in the relevant dataframe
            #             self.__user_added_times[i][1] = relevant_df['time'].iloc[-1]
            #
            #         # If it only contains the end then it must be offset too low or is too short
            #         if contains_end and not contains_start:
            #             relevant_df = relevant_df[relevant_df['time'] <= end_time]
            #
            #             # Set the start time to match the very last time of the relevant dataframe
            #             self.__user_added_times[i][2] = relevant_df['time'].iloc[0]
            #
            #         # If it contains start and end then it's just right
            #         if contains_end and contains_start:
            #             relevant_df = relevant_df[relevant_df['time'] >= start_time]
            #             relevant_df = relevant_df[relevant_df['time'] <= end_time]
            #
            #             # This dataset has been completed so there is no need to continue
            #             self.__user_added_times[i] = None
            #
            #         # Write these into the data array
            #         if asset not in list(final_prices.keys()):
            #             final_prices[asset] = relevant_df
            #         else:
            #             final_prices[asset] = pd.concat([final_prices[asset], relevant_df])

            relevant_data = []
            for j in used_ranges:
                relevant_data.append(pd.read_csv(os.path.join(cache_folder, to_string_key([asset, j[0], j[1],
                                                                                           resolution]) + ".csv")))

            if len(relevant_data) > 0:
                final_prices[asset] = pd.concat(relevant_data)

            # If there is any data left to download do it here
            for j in negative_ranges:
                print("No cached data found for " + asset + " from: " + str(j[0]) + " to " +
                      str(j[1]) + " at a resolution of " + str(resolution) + " seconds.")
                download = self.interface.get_product_history(asset,
                                                              j[0],
                                                              j[1],
                                                              resolution)

                # Write the file but this time include very accurately the start and end times
                if self.preferences['settings']['continuous_caching']:
                    if not download.empty:
                        download.to_csv(os.path.join(cache_folder, f'{asset},'
                                                                   f'{j[0]},'
                                                                   f'{j[1]+resolution},'  # This adds resolution 
                                                                                          # back to the exported
                                                                                          # time series
                                                                   f'{resolution}.csv'),
                                        index=False)

                # Write these into the data array
                if asset not in list(final_prices.keys()):
                    final_prices[asset] = download
                else:
                    final_prices[asset] = pd.concat([final_prices[asset], download])

            # After all the negative ranges are appended, we need to sort & trim
            final_prices[asset] = final_prices[asset].sort_values(by=['time'], ignore_index=True)

            # Now make sure to just trim our times to hit the start and end times
            final_prices[asset] = final_prices[asset][final_prices[asset]['time'] >= start_time]
            final_prices[asset] = final_prices[asset][final_prices[asset]['time'] <= end_time + resolution]  # Add back

        return final_prices

        # cache_folder = self.preferences['settings']["cache_location"]
        # # Make sure the cache folder exists and read files
        # try:
        #     files = os.listdir(cache_folder)
        # except FileNotFoundError:
        #     files = []
        #     os.mkdir(cache_folder)
        #
        # available_files = []
        # for i in range(len(files)):
        #     # example file name: 'BTC-USD.1622400000.1622510793.60.csv'
        #     # Remove the .csv from each of the files: BTC-USD.1622400000.1622510793.60
        #     identifier = files[i][:-4].split(",")
        #     identifier[1] = float(identifier[1])
        #     identifier[2] = float(identifier[2])
        #     identifier[3] = float(identifier[3])
        #     available_files.append(identifier)
        #
        # assets = self.preferences['price_data']['assets']  # type: dict
        # for i in assets:
        #     identifier = [i[0], i[1], i[2], i[3]]
        #     string_identifier = to_string_key(identifier)
        #     if identifier in available_files:
        #         # Read the csv here
        #         if tuple(identifier) not in self.price_dictionary.keys():
        #             print("Including: " + string_identifier + ".csv in backtest.")
        #             self.price_dictionary[tuple(identifier)] = pd.read_csv(os.path.join(cache_folder,
        #                                                                                 string_identifier + ".csv")
        #                                                                    )
        #     else:
        #         if tuple(identifier) not in self.price_dictionary.keys():
        #             print("No exact cache exists for " + str(identifier[0]) + " from " + str(identifier[1]) + " to " +
        #                   str(identifier[2]) + " at " + str(identifier[3]) + "s resolution. Downloading...")
        #             download = self.interface.get_product_history(identifier[0], identifier[1], identifier[2],
        #                                                           identifier[3])
        #             self.price_dictionary[tuple(identifier)] = download
        #             if save:
        #                 download.to_csv(os.path.join(cache_folder, string_identifier + ".csv"), index=False)
        #
        # # Merge all the same asset ids into the same dictionary spots
        # unique_assets = {}
        # for k, v in self.price_dictionary.items():
        #     if k[0] in unique_assets:
        #         unique_assets[k[0]] = pd.concat([unique_assets[k[0]], v], ignore_index=True)
        #     else:
        #         unique_assets[k[0]] = v
        #
        # return unique_assets

    def add_prices(self, asset_id, start_time, end_time, resolution, save=False):
        # Create its unique identifier
        identifier = [asset_id, int(start_time), int(end_time), int(resolution)]

        # If it's not loaded then write it to the file
        if tuple(identifier) not in self.price_dictionary.keys():
            # self.preferences['price_data']['assets'].append(identifier)
            self.__user_added_times.append(identifier)
            if save:
                self.queue_backtest_write = True

            # This makes sure that we keep track of our bounds which is just generally kind of useful
            # Any time a new price event is added we check this
            if self.user_start is None:
                self.user_start = start_time
            else:
                if start_time < self.user_start:
                    self.user_start = start_time

            if self.user_stop is None:
                self.user_stop = end_time
            else:
                if end_time > self.user_stop:
                    self.user_stop = end_time

            # Now keep track of the smallest price event added
            if self.min_resolution is None:
                self.min_resolution = resolution
            else:
                if resolution < self.min_resolution:
                    self.min_resolution = resolution
        else:
            print("already identified")

    def append_backtest_price_event(self, callback: typing.Callable, asset_id, time_interval, state_object, ohlc, init,
                                    teardown):
        if isinstance(time_interval, str):
            time_interval = time_interval_to_seconds(time_interval)
        self.price_events.append([callback, asset_id, time_interval, state_object, ohlc, init, teardown])

    def write_setting(self, key, value, save=False):
        """
        Write a setting to the .json preferences

        Args:
            key: Key under settings
            value: Value to set that settings key to
            save: Write this new setting to the file
        """
        self.preferences['settings'][key] = value
        if save:
            self.queue_backtest_write = True

    def write_initial_price_values(self, account_dictionary):
        """
        Write in a new price dictionary for the paper trade exchange.
        """
        self.interface.override_local_account(account_dictionary)

    def format_account_data(self, local_time) -> typing.Tuple[typing.Dict[typing.Union[str, typing.Any],
                                                                          typing.Union[int, typing.Any]],
                                                              typing.Dict[typing.Union[str, typing.Any],
                                                                          typing.Union[int, typing.Any]]]:

        # This is done so that only traded assets are evaluated.
        true_available = {}
        assets = self.__traded_assets
        true_account = {}
        for i in assets:
            # Grab the account status
            true_account[i] = self.interface.get_account(i)

        # Create an account total value
        value_total = 0

        no_trade_available = {}
        # No trade account total
        no_trade_value = 0

        # Save this up front so that it can be removed from the price calculation (it's always a value of 1 anyway)
        quote_value = true_account[self.quote_currency]['available'] + true_account[self.quote_currency]['hold']
        try:
            del true_account[self.quote_currency]
        except KeyError:
            pass

        for i in list(true_account.keys()):
            # Funds on hold are still added
            true_available[i] = true_account[i]['available'] + true_account[i]['hold']
            no_trade_available[i] = self.initial_account[i]['available'] + self.initial_account[i]['hold']
            currency_pair = i

            # Convert to quote (this could be optimized a bit)
            if self.__exchange_type != 'alpaca':
                currency_pair += '-'
                currency_pair += self.quote_currency

            # Get price at time
            try:
                price = self.interface.get_price(currency_pair)
            except KeyError:
                # Must be a currency we have no data for
                price = 0
            value_total += price * true_available[i]
            no_trade_value += price * no_trade_available[i]

        # Make sure to add the time key in
        true_available['time'] = local_time
        no_trade_available['time'] = local_time

        value_total += quote_value
        true_available[self.quote_currency] = quote_value

        no_trade_value += self.initial_account[self.quote_currency]['available'] + \
                          self.initial_account[self.quote_currency]['hold']

        true_available['Account Value (' + self.quote_currency + ')'] = value_total

        no_trade_available['Account Value (No Trades)'] = no_trade_value

        return true_available, no_trade_available

    def __account_was_used(self, column) -> bool:
        show_zero_delta = self.preferences['settings']['show_tickers_with_zero_delta']

        # Just check if it's in the traded assets or if the zero delta is enabled
        used = self.__traded_assets
        is_used = column in used or 'Account Value (' + self.quote_currency + ')' == column

        # Return true if they are not the same or the setting is set to true
        output = is_used or show_zero_delta
        return output

    def __next_color(self):
        # This should be a generator, but it doesn't work without doing a foreach loop
        try:
            return next(self.__color_generator)
        except StopIteration:
            self.__color_generator = Category10_10.__iter__()
            return next(self.__color_generator)

    def run(self) -> BacktestResult:
        """
        Setup
        """
        # This is where we begin logging the backtest time
        start_clock = time.time()

        # Create this initial so that we can compare how our strategy performs
        self.initial_account = self.interface.get_account()

        # Write our queued edits to the file
        if self.queue_backtest_write:
            write_backtest_preferences(self.preferences, self.backtest_settings_path)

        # TODO the preferences should really be reloaded here so that micro changes such as the quote currency reset
        #  don't need to happen for every single key type
        self.quote_currency = self.preferences['settings']['quote_account_value_in']

        # Get the symbol used for the benchmark
        benchmark_symbol = self.preferences["settings"]["benchmark_symbol"]

        if benchmark_symbol is not None:
            # Check locally for the data and add to price_cache if we do not have it
            self.add_prices(benchmark_symbol, self.user_start, self.user_stop, self.min_resolution)

        prices = self.sync_prices()

        # Organize each price into this structure: [epoch, "BTC-USD", price, open, high, low, close, volume]
        use_price = self.preferences['settings']['use_price']
        self.use_price = use_price

        self.pd_prices = {**prices}

        for k, v in prices.items():
            frame = v  # type: pd.DataFrame

            # Be sure to push these initial prices to the strategy
            try:
                self.interface.receive_price(k, v[use_price].iloc[0])
            except IndexError:
                def check_if_any_column_has_prices(price_dict: dict) -> bool:
                    """
                    In dictionary of symbols, check if at least one key has data
                    """
                    for j in price_dict:
                        if not price_dict[j].empty:
                            return True
                    return False

                if not check_if_any_column_has_prices(prices):
                    raise IndexError('No cached or downloaded data available. Try adding arguments such as to="1y" '
                                     'in the backtest command. If there should be data downloaded, try deleting your'
                                     ' ./price_caches folder.')
                else:
                    raise IndexError(f"Data for symbol {k} is empty. Are you using a symbol that is incompatible "
                                     f"with this exchange?")

            # Be sure to send in the initial time
            self.interface.receive_time(v['time'].iloc[0])

            for index, row in frame.iterrows():
                # TODO iterrows() is allegedly pretty slow
                self.prices.append([row.time, k, row[use_price],
                                    row['open'],  # (index) 3
                                    row['high'],  # 4
                                    row['low'],  # 5
                                    row['close'],  # 6
                                    row['volume']])  # 7

        # pushing these prices together makes the time go weird
        self.prices = sorted(self.prices)

        if prices == {} or self.price_events == []:
            raise ValueError("Either no price data or backtest events given. "
                             "Try setting an argument such as to='1y' in the .backtest() command.\n"
                             "Example: strategy.backtest(to='1y')")

        self.current_time = self.prices[0][0]

        self.initial_time = self.current_time

        # Add a section to the price events which controls the next time they run & change to array of dicts
        for i in range(len(self.price_events)):
            self.price_events[i] = {
                'function': self.price_events[i][0],
                'asset_id': self.price_events[i][1],
                'interval': self.price_events[i][2],
                'state_object': self.price_events[i][3],
                'next_run': self.initial_time,
                'ohlc': self.price_events[i][4],
                'init': self.price_events[i][5],
                'teardown': self.price_events[i][6]
            }

        # Initialize this before the callbacks so it works in the initialization functions
        self.time = self.initial_time

        # Turn on backtesting immediately after setting the time
        self.interface.set_backtesting(True)

        # Run the initialization functions for the price events
        print("\nInitializing...")
        for i in self.price_events:
            if i['init'] is not None:
                i['init'](i['asset_id'], i['state_object'])

        """
        Begin backtesting
        """

        # Re-evaluate the traded assets account
        # This is mainly used if the user has an account with some value that gets added in at the backtest point
        # This occurs after initialization so there has to be a function to test & re-evaluate that
        self.interface.evaluate_traded_account_assets()
        column_keys = list.copy(self.interface.traded_assets)

        # Comically if you don't include the quote at any point there will be an error
        if self.quote_currency not in column_keys:
            column_keys.append(self.quote_currency)
        # If they start a price event on something they don't own, this should also be included
        for i in self.price_events:
            base_asset = get_base_asset(i['asset_id'])
            quote_asset = get_quote_asset(i['asset_id'])
            if base_asset not in column_keys:
                column_keys.append(base_asset)
            if quote_asset not in column_keys:
                column_keys.append(quote_asset)
        self.__traded_assets = list.copy(column_keys)
        column_keys.append('time')

        # column_keys = ['time']
        # for i in account.keys():
        #     column_keys.append(i['currency'])

        cycle_status = pd.DataFrame(columns=column_keys)

        no_trade_cycle_status = pd.DataFrame(columns=column_keys)

        # Append dictionaries to this to make the pandas dataframe
        price_data = []

        # Append dictionaries to this to make the no trade dataframe
        no_trade = []

        # Add an initial account row here
        if self.preferences['settings']['save_initial_account_value']:
            available_dict, no_trade_dict = self.format_account_data(self.initial_time)
            price_data.append(available_dict)
            no_trade.append(no_trade_dict)

        show_progress = self.preferences['settings']['show_progress_during_backtest']

        ignore_exceptions = self.preferences['settings']['ignore_user_exceptions']

        print("\nBacktesting...")
        price_number = len(self.prices)

        # This dictionary will hold price arrays below sorted by symbol which will allow us to grab most
        #  recent ohlcv data when necessary
        ohlcv_tracker = {}

        try:
            for i in range(price_number):
                #                 row.time,      k,  use_price, 'open', 'high', 'low','close','volume'
                # Formatted like [1609146000.0, 'AAPL', 132.99, 132.72, 133.0, 132.6, 132.99, 32603.0]
                price_array = self.prices[i]
                if show_progress:
                    if i % 100 == 0:
                        update_progress(i / price_number)
                self.interface.receive_price(asset_id=price_array[1], new_price=price_array[2])
                self.current_time = price_array[0]
                self.interface.evaluate_limits()

                # This will keep the most recent price event data organized by symbol
                ohlcv_tracker[price_array[1]] = price_array

                while True:
                    # Need to go through and establish an order for each of the price events
                    self.price_events = sorted(self.price_events, key=lambda sort_key: sort_key['next_run'])

                    # Now the lowest one has to go past the current time to be invalid
                    if self.price_events[0]['next_run'] > self.current_time:
                        break

                    local_time = self.price_events[0]['next_run']

                    # Export the time for strategy
                    self.time = local_time

                    self.interface.receive_time(local_time)

                    # This is the actual callback to the user space
                    try:
                        if self.price_events[0]['ohlc']:
                            # This pulls all the price data out of the price array defined on line 260
                            ohlcv_array = ohlcv_tracker[self.price_events[0]['asset_id']]
                            self.price_events[0]['function']({'open': ohlcv_array[3],
                                                              'high': ohlcv_array[4],
                                                              'low': ohlcv_array[5],
                                                              'close': ohlcv_array[6],
                                                              'volume': ohlcv_array[7]},

                                                             self.price_events[0]['asset_id'],
                                                             self.price_events[0]['state_object'])
                        else:
                            self.price_events[0]['function'](self.interface.get_price(self.price_events[0]['asset_id']),
                                                             self.price_events[0]['asset_id'], self.price_events[0][
                                                                 'state_object'])
                    except Exception as e:
                        if ignore_exceptions:
                            traceback.print_exc()
                        else:
                            raise e

                    # Delay the next run until after the interval
                    self.price_events[0]['next_run'] += self.price_events[0]['interval']

                    available_dict, no_trade_dict = self.format_account_data(local_time)

                    price_data.append(available_dict)

                    no_trade.append(no_trade_dict)

            # Finish filling the progress bar
            if show_progress:
                update_progress(1)

            # Finally, run the teardown functions
            for i in self.price_events:
                # Pull the teardown and pass the state object
                if callable(i['teardown']):
                    i['teardown'](i['state_object'])
        except Exception:
            traceback.print_exc()

        # Reset time to be None to indicate we're no longer in a backtest
        self.time = None

        # Push the accounts to the dataframe
        cycle_status = pd.concat([cycle_status, pd.DataFrame(price_data)], ignore_index=True).sort_values(by=['time'])

        if len(cycle_status) == 0:
            raise RuntimeError("Empty result - no valid backtesting events occurred. Was there an error?.")

        no_trade_cycle_status = pd.concat([no_trade_cycle_status, pd.DataFrame(no_trade)], ignore_index=True)\
            .sort_values(by=['time'])

        def is_number(s):
            try:
                float(s)
                # Love how bools cast to a number
                return not isinstance(s, bool)
            except ValueError:
                return False

        history_and_returns = {
            'history': cycle_status
        }
        metrics_indicators = {}
        user_callbacks = {}

        result_object = BacktestResult(history_and_returns, {
            'created': self.interface.paper_trade_orders,
            'limits_executed': self.interface.executed_orders,
            'limits_canceled': self.interface.canceled_orders,
            'executed_market_orders': self.interface.market_order_execution_details
        }, self.pd_prices, self.initial_time, self.interface.time(), self.quote_currency, self.price_events, [])

        # If they set resampling we use resampling for everything
        resample_setting = self.preferences['settings']['resample_account_value_for_metrics']
        if isinstance(resample_setting, str) or is_number(resample_setting):
            resample_to = resample_setting
        else:
            info_print('Resampling value not set, defaulting to 1 day.')
            resample_to = '1d'

        interval_value = time_interval_to_seconds(resample_to)

        # This is where we run the actual resample
        resampled_account_data_frame = result_object.resample_account('Account Value (' + self.quote_currency + ')',
                                                                      interval_value)

        history_and_returns['resampled_account_value'] = resampled_account_data_frame

        returns = resampled_account_data_frame.copy(deep=True)

        # Default diff parameters should do it
        returns['value'] = returns['value'].pct_change()

        # Now write it to our dictionary
        history_and_returns['returns'] = returns

        # -----=====*****=====-----
        metrics_indicators['Compound Annual Growth Rate (%)'] = metrics.cagr(history_and_returns)
        try:
            metrics_indicators['Cumulative Returns (%)'] = metrics.cum_returns(history_and_returns)
        except ZeroDivisionError as e_:
            metrics_indicators['Cumulative Returns (%)'] = f'failed: {e_}'

        def attempt(math_callable: typing.Callable, dict_of_dataframes: dict, kwargs: dict = None):
            try:
                if kwargs is None:
                    kwargs = {}
                result = math_callable(dict_of_dataframes, **kwargs)
                if result == np.NAN:
                    result = None
                return result
            except (ZeroDivisionError, Exception) as e__:
                return f'failed: {e__}'

        risk_free_return_rate = self.preferences['settings']["risk_free_return_rate"]
        metrics_indicators['Max Drawdown (%)'] = attempt(metrics.max_drawdown, history_and_returns)
        metrics_indicators['Variance (%)'] = attempt(metrics.variance, history_and_returns,
                                                     {'trading_period': interval_value})
        metrics_indicators['Sortino Ratio'] = attempt(metrics.sortino, history_and_returns,
                                                      {'risk_free_rate': risk_free_return_rate,
                                                       'trading_period': interval_value})
        metrics_indicators['Sharpe Ratio'] = attempt(metrics.sharpe, history_and_returns,
                                                     {'risk_free_rate': risk_free_return_rate,
                                                      'trading_period': interval_value})
        metrics_indicators['Calmar Ratio'] = attempt(metrics.calmar, history_and_returns,
                                                     {'trading_period': interval_value})
        metrics_indicators['Volatility'] = attempt(metrics.volatility, history_and_returns,
                                                   {'trading_period': interval_value})
        metrics_indicators['Value-at-Risk'] = attempt(metrics.var, history_and_returns)
        metrics_indicators['Conditional Value-at-Risk'] = attempt(metrics.cvar, history_and_returns)
        
        # Add risk-free-return rate to dictionary
        metrics_indicators['Risk Free Return Rate'] = risk_free_return_rate
        # metrics_indicators['beta'] = attempt(metrics.beta, dataframes)
        # Add the interval value to dictionary
        metrics_indicators['Resampled Time'] = interval_value
        # -----=====*****=====-----

        # If a benchmark was requested, add it to the pd_prices frame
        if benchmark_symbol is not None:
            # Resample the benchmark results
            resampled_benchmark_value = result_object.resample_account(benchmark_symbol,
                                                                       interval_value,
                                                                       use_asset_history=True,
                                                                       use_price=use_price)
            
            # Push data into the dictionary for use by the metrics
            history_and_returns['benchmark_value'] = resampled_benchmark_value
            history_and_returns['benchmark_returns'] = resampled_benchmark_value.copy(deep=True)
            history_and_returns['benchmark_returns']['value'] = \
                history_and_returns['benchmark_returns']['value'].pct_change()
            
            # Calculate beta
            metrics_indicators['Beta'] = attempt(metrics.beta, history_and_returns, 
                                                 {"trading_period": interval_value})

        # This trys to reference vars created in the resample portion, so it has to also be in the specified
        #  resample if
        # --
        # Run this last so that the user can override what they want
        for callback in self.callbacks:
            user_callbacks[callback.__name__] = callback(history_and_returns, metrics_indicators)

        # Remove NaN values here
        history_and_returns['resampled_account_value'] = history_and_returns['resampled_account_value'].\
            where(history_and_returns['resampled_account_value'].notnull(), None)

        # Remove NaN values on this one too
        history_and_returns['returns'] = history_and_returns['returns'].where(history_and_returns['returns'].notnull(),
                                                                              None)
        # Lastly remove Nan values in the metrics
        for i in metrics_indicators:
            if not isinstance(metrics_indicators[i], str) and np.isnan(metrics_indicators[i]):
                metrics_indicators[i] = None

        # Assign all these new values back to the result object
        result_object.history_and_returns = history_and_returns
        result_object.metrics = metrics_indicators
        result_object.user_callbacks = user_callbacks

        figures = []
        # This modifies the platform result in place
        platform_result = format_platform_result(result_object)
        if self.preferences['settings']['GUI_output']:
            def internal_backtest_viewer():
                # for i in self.prices:
                #     result_index = cycle_status['time'].sub(i[0]).abs().idxmin()
                #     for i in cycle_status.iloc[result_index]:

                hover = HoverTool(
                    tooltips=[
                        ('value', '@value')
                    ],

                    # formatters={
                    #     'time': 'datetime',  # use 'datetime' formatter for 'date' field
                    #     '@{value}': 'printf',   # use 'printf' formatter for '@{adj close}' field
                    # },

                    # display a tooltip whenever the cursor is vertically in line with a glyph
                    mode='vline'
                )

                # Define a helper function to avoid repeating code
                def add_trace(self_, figure_, time_, data_, label):
                    source = ColumnDataSource(data=dict(
                        time=time_,
                        value=data_.values.tolist()
                    ))
                    figure_.step('time', 'value',
                                 source=source,
                                 line_width=2,
                                 color=self_.__next_color(),
                                 legend_label=label,
                                 mode="after")

                global_x_range = None

                time = [dt.fromtimestamp(ts) for ts in cycle_status['time']]

                for column in cycle_status:
                    if column != 'time' and self.__account_was_used(column):
                        p = figure(plot_width=900, plot_height=200, x_axis_type='datetime')
                        add_trace(self, p, time, cycle_status[column], column)

                        # Add the no-trade line to the backtest
                        if column == 'Account Value (' + self.quote_currency + ')':
                            add_trace(self, p, time, no_trade_cycle_status['Account Value (No Trades)'],
                                      'Account Value (No Trades)')

                            # Add the benchmark, if requested
                            if benchmark_symbol is not None:
                                # This normalizes the benchmark value
                                initial_account_value = cycle_status['Account Value (' +
                                                                     self.quote_currency + ')'].iloc[0]
                                initial_benchmark_value = prices[benchmark_symbol][use_price].iloc[0]

                                # This multiplier brings the initial asset price to the initial account value
                                # initial_account_value = initial_benchmark_value * x
                                multiplier = initial_account_value / initial_benchmark_value

                                normalized_compare_series = prices[benchmark_symbol][use_price].multiply(multiplier)
                                normalized_compare_time_series = prices[benchmark_symbol]['time']

                                # We need to also cast the time series that is needed to compare
                                # because it's only been done for the cycle status time
                                normalized_compare_time_series = [dt.fromtimestamp(ts) for ts in
                                                                  normalized_compare_time_series]
                                add_trace(self, p, normalized_compare_time_series,
                                          normalized_compare_series,
                                          f'Normalized Benchmark ({benchmark_symbol})')

                        p.add_tools(hover)

                        # Format graph
                        p.legend.location = "top_left"
                        p.legend.title = column
                        p.legend.title_text_font_style = "bold"
                        p.legend.title_text_font_size = "20px"
                        if global_x_range is None:
                            global_x_range = p.x_range
                        else:
                            p.x_range = global_x_range

                        figures.append(p)

                show(bokeh_columns(figures))
                info_print(f'Make an account to take advantage of the platform backtest viewer: '
                           f'https://app.blankly.finance/5Z9MWfnUzwIyy9Qv385a/1Ss7zybwN8aMAbWb3lSH/'
                           f'aG3LE1LzHnY24oqtBMS3/backtest')

            # This is where we end the backtesting time
            stop_clock = time.time()

            try:
                json_file = json.loads(open('./blankly.json').read())
                api_key = json_file['api_key']
                api_pass = json_file['api_pass']
                # Need this to generate the URL
                # Need this to know where to post to
                model_id = json_file['model_id']

                requests.post(f'https://events.blankly.finance/v1/backtest/result', json=platform_result, headers={
                    'api_key': api_key,
                    'api_pass': api_pass,
                    'model_id': model_id
                })

                requests.post(f'https://events.blankly.finance/v1/backtest/status', json={
                    'successful': True,
                    'status_summary': 'Completed',
                    'status_details': '',
                    'time_elapsed': stop_clock-start_clock,
                    'backtest_id': platform_result['backtest_id']
                }, headers={
                    'api_key': api_key,
                    'api_pass': api_pass,
                    'model_id': model_id
                })

                import webbrowser

                link = f'https://app.blankly.finance/{api_key}/{model_id}/{platform_result["backtest_id"]}' \
                       f'/backtest'
                webbrowser.open(
                    link
                )
                info_print(f'View your backtest here: {link}')
            except (FileNotFoundError, KeyError):
                internal_backtest_viewer()

        # Finally, write the figures in
        result_object.figures = figures

        self.interface.set_backtesting(False)
        return result_object



"""
    Oanda interface custom unit tests.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
from pathlib import Path

import pytest

import blankly
from blankly.exchanges.interfaces.oanda.oanda_interface import OandaInterface
from tests.helpers.comparisons import validate_response
from tests.testing_utils import forex_market_open


@pytest.fixture
def oanda_interface():
    keys_file_path = Path("tests/config/keys.json").resolve()
    settings_file_path = Path("tests/config/settings.json").resolve()

    oanda = blankly.Oanda(keys_path=keys_file_path,
                          settings_path=settings_file_path,
                          portfolio_name='oanda test portfolio')

    # auth_obj = OandaAuth(str(keys_file_path), "oanda test portfolio")
    # _, oanda_interface = DirectCallsFactory.create("oanda", auth_obj, str(settings_file_path))
    return oanda.interface


def test_get_exchange(oanda_interface: OandaInterface) -> None:
    assert oanda_interface.get_exchange_type() == 'oanda'


def test_get_price(oanda_interface: OandaInterface) -> None:
    assert isinstance(oanda_interface.get_price("EUR-USD"), float)
    assert isinstance(oanda_interface.get_price("ZAR-JPY"), float)


def test_get_cash(oanda_interface: OandaInterface) -> None:
    assert isinstance(oanda_interface.cash, float)


def test_marketorder_comprehensive(oanda_interface: OandaInterface) -> None:
    # query for the unique ID of EUR_USD
    if not forex_market_open():
        return

    products = oanda_interface.get_products()

    found_eur_usd = False
    for product in products:
        if product['symbol'] == 'EUR-USD':
            found_eur_usd = True

    assert found_eur_usd

    eur_usd = 'EUR-USD'
    market_buy_order = oanda_interface.market_order(eur_usd, 'buy', 200)

    time.sleep(2)
    resp = oanda_interface.get_order(eur_usd, market_buy_order.get_id())

    # verify resp is correct
    validate_response(oanda_interface.needed['market_order'], resp)

    # Todo: validate market sell order object can query its info correctly
    oanda_interface.market_order(eur_usd, 'sell', 200)

    time.sleep(2)
    resp = oanda_interface.get_order(eur_usd, market_buy_order.get_id())

    # verify resp is correct
    validate_response(oanda_interface.needed['market_order'], resp)


def test_get_products(oanda_interface: OandaInterface) -> None:
    products = oanda_interface.get_products()
    for product in products:
        validate_response(oanda_interface.needed['get_products'], product)


def test_get_account(oanda_interface: OandaInterface) -> None:
    account = oanda_interface.get_account()
    usd_found = False
    for key, val in account.items():
        validate_response(oanda_interface.needed['get_account'], val)
        if key == "USD":
            usd_found = True

    assert usd_found

    account = oanda_interface.get_account('USD')
    assert 'available' in account
    assert 'hold' in account
    assert isinstance(account['available'], float)
    assert isinstance(account['hold'], float)


def test_limitorder_comprehensive(oanda_interface: OandaInterface) -> None:
    eur_usd = 'EUR-USD'
    limit_buy_order = oanda_interface.limit_order(eur_usd, 'buy', 1, 5)

    time.sleep(1)
    resp = oanda_interface.get_order(eur_usd, limit_buy_order.get_id())

    validate_response(oanda_interface.needed['limit_order'], resp)

    # now cancel the order
    resp = oanda_interface.cancel_order(eur_usd, limit_buy_order.get_id())
    validate_response(oanda_interface.needed['cancel_order'], resp)


def test_get_order(oanda_interface: OandaInterface) -> None:
    locally_created_orders = []
    for i in range(5):
        locally_created_orders.append(oanda_interface.limit_order("EUR-USD", 'buy', 1, 5))
        time.sleep(0.1)

    orders = oanda_interface.get_open_orders()
    for order in orders:
        needed = oanda_interface.choose_order_specificity(order['type'])
        validate_response(needed, order)

    orders = oanda_interface.get_open_orders("EUR-USD")
    for order in orders:
        needed = oanda_interface.choose_order_specificity(order['type'])
        validate_response(needed, order)

    for order in locally_created_orders:
        resp = oanda_interface.cancel_order("EUR-USD", order.get_id())
        validate_response(oanda_interface.needed['cancel_order'], resp)


def test_get_filters(oanda_interface: OandaInterface) -> None:
    products = oanda_interface.get_products()
    for product in products:
        resp = oanda_interface.get_order_filter(product['symbol'])
        validate_response(oanda_interface.needed['get_order_filter'], resp)


def test_get_product_history(oanda_interface: OandaInterface) -> None:
    pass


def test_history(oanda_interface: OandaInterface) -> None:
    pass

#
# def test_get_all_open_orders(oanda_interface: OandaInterface) -> None:
#     start = dateparser.parse("2021-02-04 9:30AM EST").timestamp()
#     end = dateparser.parse("2021-02-04 9:35AM EST").timestamp()
#
#     # bars = oanda_interface.history("EUR_USD", to=5, resolution=60)
#     # bars = oanda_interface.history("EUR_USD", to=5, resolution=60, end_date=end)
#     bars = oanda_interface.history("EUR_USD", to=5, resolution=60*3, end_date=end)
#     assert False
#



"""
    Keyless interface definition
    Copyright (C) 2022 Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import pandas as pd
from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.utils.utils import AttributeDict, get_base_asset, get_quote_asset, add_all_products


# This just happens to also inherit from the exchange interface
class KeylessAPI(ExchangeInterface):
    def __init__(self, dataset):

        self.__dataset: pd.DataFrame = pd.read_csv(dataset)

        self.__products = []
        self.__accounts = {}

        super().__init__('keyless', self)

    def init_exchange(self):
        for i in self.__dataset.columns:
            if i == "time":
                continue
            else:
                self.__products.append({
                    'symbol': i
                })

        # Now pull out the single assets
        self.__accounts = add_all_products({}, self.__products)

    def __invalid_live(self):
        raise RuntimeError("Cannot use a keyless exchange in live trading. Please insert keys and begin using exchanges"
                           " such as blankly.Alpaca() or blankly.CoinbasePro().")

    def get_exchange_type(self):
        return 'keyless'

    def get_account(self, symbol: str = None) -> AttributeDict:
        return AttributeDict(self.__accounts)

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        # Ignore resolution because the backtest engine will deal with that for us
        symbol_series = self.__dataset[symbol]
        time_series = self.__dataset['time']

        symbol_series = symbol_series[time_series >= epoch_start]
        symbol_series = symbol_series[time_series <= epoch_stop]

        time_series = time_series[time_series >= epoch_start]
        time_series = time_series[time_series <= epoch_stop]

        df = pd.DataFrame({
            'open': symbol_series,
            'high': symbol_series,
            'low': symbol_series,
            'close': symbol_series,
            'volume': symbol_series,
            'time': time_series
        })
        return df

    def get_products(self) -> list:
        return self.__products

    def market_order(self, symbol: str, side: str, size: float):
        self.__invalid_live()

    def limit_order(self, symbol: str, side: str, price: float, size: float):
        self.__invalid_live()

    def cancel_order(self, symbol: str, order_id: str):
        self.__invalid_live()

    def get_open_orders(self, symbol: str = None):
        self.__invalid_live()

    def get_order(self, symbol: str, order_id: str):
        self.__invalid_live()

    def get_fees(self) -> dict:
        return {
            'maker_fee_rate': .005,
            'taker_fee_rate': .005
        }

    def get_order_filter(self, symbol: str) -> dict:
        return {
            "symbol": symbol,
            "base_asset": get_base_asset(symbol),
            "quote_asset": get_quote_asset(symbol),
            "max_orders": 1000000000000000,
            "limit_order": {
                "base_min_size": 0,  # Minimum size to buy
                "base_max_size": 1000000000000000,  # Maximum size to buy
                "base_increment": 0,  # Specifies the minimum increment
                # for the base_asset.
                "price_increment": 0,

                "min_price": 0,
                "max_price": 1000000000000000,
            },
            'market_order': {
                "fractionable": True,

                "base_min_size": 0,  # Minimum size to buy
                "base_max_size": 1000000000000000,  # Maximum size to buy
                "base_increment": 0,  # Specifies the minimum increment

                "quote_increment": 0,  # Specifies the min order price as well
                # as the price increment.
                "buy": {
                    "min_funds": 0,
                    "max_funds": 1000000000000000,
                },
                "sell": {
                    "min_funds": 0,
                    "max_funds": 1000000000000000,
                },
            },
            "exchange_specific": {}
        }

    def get_price(self, symbol: str, time=None):
        self.__invalid_live()



"""
    Class to manage the orderbook, adding, removing and updating - as well as provide user interaction
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import traceback
import warnings
from typing import List

import requests

import blankly.exchanges.auth.utils
import blankly.utils.utils
from blankly.exchanges.interfaces.alpaca.alpaca_websocket import Tickers as Alpaca_Websocket
from blankly.exchanges.interfaces.binance.binance_websocket import Tickers as Binance_Orderbook
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_websocket import Tickers as Coinbase_Pro_Orderbook
from blankly.exchanges.managers.websocket_manager import WebsocketManager


def sort_list_tuples(list_with_tuples: list) -> List[tuple]:
    return sorted(list_with_tuples, key=lambda x: x[0])


def binance_snapshot(symbol, limit):
    # TODO this should be using the REST set that Arun puts in
    params = {
        "symbol": symbol,
        "limit": limit
    }
    response = requests.get("https://api.binance.com/api/v3/depth", params=params).json()
    try:
        buys_response = response['bids']
    except KeyError:
        raise KeyError(response)
    sells_response = response['asks']
    buys = []
    sells = []
    for i in buys_response:
        buys.append((float(i[0]), float(i[1])))
    for i in sells_response:
        sells.append((float(i[0]), float(i[1])))

    buys = sort_list_tuples(buys)
    sells = sort_list_tuples(sells)
    print("Orderbook snapshot acquired for: " + symbol.upper())
    return buys, sells


def remove_price(book: list, price: float) -> list:
    for j in range(len(book)):
        price_at_index = book[j][0]
        if price_at_index > price:
            break
        if price_at_index == price:
            book.pop(j)
            break

    return book


class OrderbookManager(WebsocketManager):
    def __init__(self, default_exchange, default_symbol):
        """
        Create a new orderbook manager
        Args:
            default_exchange: Add an exchange name for the manager to favor
            default_symbol: Add a default currency for the manager to favor
        """
        self.__default_exchange = default_exchange
        self.__default_currency = default_symbol

        self.__orderbooks = {default_exchange: {}}

        self.__websockets = {default_exchange: {}}

        self.__websockets_callbacks = {default_exchange: {}}

        self.__websockets_kwargs = {default_exchange: {}}

        # Create the abstraction for adding many managers
        super().__init__(self.__websockets, default_symbol, default_exchange)

    def create_orderbook(self, callback,
                         override_symbol=None,
                         override_exchange=None,
                         initially_stopped=False,
                         **kwargs):
        """
        Create an orderbook for a given exchange
        Args:
            callback: Callback object for the function. Should be something like self.price_event
            override_symbol: Override the default currency id
            override_exchange: Override the default exchange
            initially_stopped: Keep the websocket stopped when created
            kwargs: Add any other parameters that should be passed into a callback function to identify
                it or modify behavior
        """

        use_sandbox = self.preferences['settings']['use_sandbox_websockets']

        exchange_name = self.__default_exchange
        # Ensure the ticker dict has this overridden exchange
        if override_exchange is not None:
            if override_exchange not in self.__websockets.keys():
                self.__websockets[override_exchange] = {}
                self.__websockets_callbacks[override_exchange] = {}
                self.__websockets_kwargs[override_exchange] = {}
            # Write this value so it can be used later
            exchange_name = override_exchange

        # Ensure that we always have a key the relevant orderbook
        if exchange_name not in self.__orderbooks:
            self.__orderbooks[exchange_name] = {}

        if exchange_name == "coinbase_pro":
            if override_symbol is None:
                override_symbol = self.__default_currency

            if use_sandbox:
                websocket = Coinbase_Pro_Orderbook(override_symbol, "level2",
                                                   pre_event_callback=self.coinbase_snapshot_update,
                                                   initially_stopped=initially_stopped,
                                                   WEBSOCKET_URL="wss://ws-feed-public.sandbox.pro.coinbase.com")
            else:
                websocket = Coinbase_Pro_Orderbook(override_symbol, "level2",
                                                   pre_event_callback=self.coinbase_snapshot_update,
                                                   initially_stopped=initially_stopped
                                                   )
            # This is where the sorting magic happens
            websocket.append_callback(self.coinbase_update)

            # Store this object
            self.__websockets['coinbase_pro'][override_symbol] = websocket
            self.__websockets_callbacks['coinbase_pro'][override_symbol] = [callback]
            self.__websockets_kwargs['coinbase_pro'][override_symbol] = kwargs
            self.__orderbooks['coinbase_pro'][override_symbol] = {
                "bids": [],
                "asks": []
            }
            return websocket
        elif exchange_name == "binance":
            if override_symbol is None:
                override_symbol = self.__default_currency

            # Lower the keys to subscribe
            specific_currency_id = blankly.utils.to_exchange_symbol(override_symbol, "binance").lower()

            if use_sandbox:
                websocket = Binance_Orderbook(specific_currency_id, "depth", initially_stopped=initially_stopped,
                                              WEBSOCKET_URL="wss://testnet.binance.vision/ws")
            else:
                websocket = Binance_Orderbook(specific_currency_id, "depth", initially_stopped=initially_stopped)

            websocket.append_callback(self.binance_update)

            # binance returns the keys in all UPPER so the books should be created based on response
            specific_currency_id = specific_currency_id.upper()
            self.__websockets['binance'][specific_currency_id] = websocket
            self.__websockets_callbacks['binance'][specific_currency_id] = [callback]
            self.__websockets_kwargs['binance'][specific_currency_id] = kwargs

            buys, sells = binance_snapshot(specific_currency_id, 1000)
            self.__orderbooks['binance'][specific_currency_id] = {
                "bids": buys,
                "asks": sells
            }

        elif exchange_name == "alpaca":
            warning_string = "Alpaca only allows the viewing of the bid/ask spread, not a total orderbook."
            warnings.warn(warning_string)
            if override_symbol is None:
                override_symbol = self.__default_currency

            stream = self.preferences['settings']['alpaca']['websocket_stream']
            override_symbol = blankly.utils.to_exchange_symbol(override_symbol, "alpaca")

            if use_sandbox:
                websocket = Alpaca_Websocket(override_symbol, 'quotes', initially_stopped=initially_stopped,
                                             WEBSOCKET_URL=
                                             "wss://paper-api.alpaca.markets/stream/v2/{}/".format(stream))
            else:
                websocket = Alpaca_Websocket(override_symbol, 'quotes', initially_stopped=initially_stopped,
                                             WEBSOCKET_URL="wss://stream.data.alpaca.markets/v2/{}/".format(stream))

            websocket.append_callback(self.alpaca_update)

            self.__websockets['alpaca'][override_symbol] = websocket
            self.__websockets_callbacks['alpaca'][override_symbol] = [callback]
            self.__websockets_kwargs['alpaca'][override_symbol] = kwargs

            self.__orderbooks['alpaca'][override_symbol] = {
                "bids": [],
                "asks": [],
            }

        else:
            print(exchange_name + " ticker not supported, skipping creation")

    def coinbase_snapshot_update(self, update):
        print("Orderbook snapshot acquired for: " + update['product_id'])
        # Clear whatever book we had
        book = {
            "bids": [],
            "asks": []
        }
        # Get all bids
        buys = update['bids']
        # Convert these to float and write to our order dictionaries
        for i in range(len(buys)):
            buy = buys[i]
            book['bids'].append((float(buy[0]), float(buy[1])))

        sells = update['asks']
        for i in range(len(sells)):
            sell = sells[i]
            book['asks'].append((float(sell[0]), float(sell[1])))

        book["bids"] = sort_list_tuples(book["bids"])
        book["asks"] = sort_list_tuples(book["asks"])

        self.__orderbooks['coinbase_pro'][update['product_id']] = book

    def coinbase_update(self, update):
        # Side is first in list
        side = update['changes'][0][0]

        # Have to convert coinbase to use this
        if side == 'buy':
            side = 'bids'
        elif side == 'sell':
            side = 'asks'

        # Price is second
        price = float(update['changes'][0][1])
        qty = float(update['changes'][0][2])
        book = self.__orderbooks['coinbase_pro'][update['product_id']][side]  # type: list

        # Quantity at that point is third
        if qty == 0:
            book = remove_price(book, price)
        else:
            book.append((price, qty))

        book = sort_list_tuples(book)
        self.__orderbooks['coinbase_pro'][update['product_id']][side] = book

        # Iterate through the callback list
        callbacks = self.__websockets_callbacks['coinbase_pro'][update['product_id']]
        for i in callbacks:
            i(self.__orderbooks['coinbase_pro'][update['product_id']],
              **self.__websockets_kwargs['coinbase_pro'][update['product_id']])

    def binance_update(self, update):
        try:
            # TODO this needs a snapshot to work correctly, which needs arun's rest code
            # Get symbol first
            symbol = update['s']

            # Get symbol for orderbook
            book_buys = self.__orderbooks['binance'][symbol]['bids']  # type: list
            book_sells = self.__orderbooks['binance'][symbol]['asks']  # type: list

            # Buys are b, count from low to high with reverse (which is the ::-1 thing)
            new_buys = update['b'][::-1]  # type: list
            for i in new_buys:
                i[0] = float(i[0])
                i[1] = float(i[1])
                if i[1] == 0:
                    book_buys = remove_price(book_buys, i[0])
                else:
                    book_buys.append((i[0], i[1]))

            # Asks are sells, these are also counted from low to high
            new_sells = update['a']  # type: list
            for i in new_sells:
                i[0] = float(i[0])
                i[1] = float(i[1])
                if i[1] == 0:
                    book_sells = remove_price(book_sells, i[0])
                else:
                    book_sells.append((i[0], i[1]))

            # Now sort them
            book_buys = sort_list_tuples(book_buys)
            book_sells = sort_list_tuples(book_sells)

            self.__orderbooks['binance'][symbol]['bids'] = book_buys
            self.__orderbooks['binance'][symbol]['asks'] = book_sells

            # Pass in this new updated orderbook
            callbacks = self.__websockets_callbacks['binance'][symbol]
            for i in callbacks:
                i(self.__orderbooks['binance'][symbol],
                  **self.__websockets_kwargs['binance'][symbol])
        except Exception:
            traceback.print_exc()

    def alpaca_update(self, update: dict):
        # Alpaca only gives the spread, no orderbook depth (alpaca is very bad)
        symbol = update['S']
        self.__orderbooks['alpaca'][symbol]['bids'] = [(update['bp'], update['bs'])]
        self.__orderbooks['alpaca'][symbol]['asks'] = [(update['ap'], update['as'])]

        callbacks = self.__websockets_callbacks['alpaca'][symbol]
        for i in callbacks:
            i(self.__orderbooks['alpaca'][symbol],
              **self.__websockets_kwargs['alpaca'][symbol])

    def append_orderbook_callback(self, callback_object, override_symbol=None, override_exchange=None):
        """
        These are appended calls to a sorted orderbook. Functions added to this will be fired every time the orderbook
        changes.
        Args:
            callback_object: Reference for the callback function. The price_event(self, tick)
                function would be passed in as just self.price_event -- no parenthesis or arguments, just the reference
            override_symbol: Ticker id, such as "BTC-USD" or exchange equivalents.
            override_exchange: Forces the manager to use a different supported exchange.
        """
        if override_symbol is None:
            override_symbol = self.__default_currency

        if override_exchange is None:
            override_exchange = self.__default_exchange

        self.__websockets_callbacks[override_exchange][override_symbol].append(callback_object)

    def get_most_recent_orderbook(self, override_symbol=None, override_exchange=None):
        """
        Get the most recent orderbook under a currency and exchange.

        Args:
            override_symbol: Ticker id, such as "BTC-USD" or exchange equivalents.
            override_exchange: Forces the manager to use a different supported exchange.
        """
        if override_symbol is None:
            override_symbol = self.__default_currency

        if override_exchange is None:
            override_exchange = self.__default_exchange

        return self.__orderbooks[override_exchange][override_symbol]



"""
    Object for cleanly storing & printing complex backtest results.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pandas import DataFrame, to_datetime, Timestamp
from blankly.utils.utils import time_interval_to_seconds as _time_interval_to_seconds, info_print


class BacktestResult:
    def __init__(self, history_and_returns: dict, trades: dict, history: dict,
                 start_time: float, stop_time: float, quote_currency: str, price_events: list, figures: list):
        # This can use a ton of memory if these attributes are not cleared
        self.history_and_returns = history_and_returns
        self.metrics = None  # Assigned after construction
        self.user_callbacks = None  # Assigned after construction
        self.trades = trades
        self.history = history

        self.quote_currency = quote_currency
        self.price_events = price_events

        self.start_time = start_time
        self.stop_time = stop_time

        self.figures = figures

    def get_account_history(self) -> DataFrame:
        return self.history_and_returns['history']

    def get_returns(self) -> DataFrame:
        return self.history_and_returns['returns']

    def get_resampled_account(self) -> DataFrame:
        return self.history_and_returns['resampled_account_value']

    def get_user_callback_results(self) -> dict:
        return self.user_callbacks

    def get_metrics(self) -> dict:
        return self.metrics

    def resample_account(self, symbol, interval: [str, float],
                         use_asset_history: bool = False,
                         use_price=None) -> DataFrame:
        """
        Resample the raw account value metrics to any resolution

        Args:
            symbol: The column to resample at the interval resolution. This can include the account value column
            interval: A string such as '1h' or '1m' or a number in seconds such as 3600 or 60 which the values
            will be resampled at
            use_asset_history: Use the history from the assets rather than the account history
            use_price: Specify a price to use when querying comparison columns
        """
        search_index = 0

        def search_price(values, times, epoch):
            # In this case because each asset is called individually
            def search(arr, size, x):
                # Use a global search index to accelerate search
                nonlocal search_index

                while True:
                    if search_index == size:
                        # Must be the last one in the list
                        return search_index - 1

                    if arr[search_index] <= x <= arr[search_index + 1]:
                        # Found it in this range
                        return search_index
                    else:
                        search_index += 1
            try:
                # Iterate and find the correct quote price
                index_ = search(times, len(times), epoch)
                return values[index_]
            except KeyError:
                # Not a currency that we have data for at all
                return 0

        resampled_array = []
        interval = _time_interval_to_seconds(interval)

        if use_asset_history:
            # Find the necessary values to assemble the resamples
            time_array = self.history[symbol]['time'].tolist()
            price_array = self.history[symbol][use_price].tolist()
        else:
            # Find the necessary values to assemble the resamples
            time_array = self.history_and_returns['history']['time'].tolist()
            price_array = self.history_and_returns['history'][symbol].tolist()

        # Add the epoch
        epoch_start = time_array[0]
        epoch_stop = time_array[-1]

        while epoch_start <= epoch_stop:
            # Append this dict to the array
            resampled_array.append({
                'time': epoch_start,
                'value': search_price(price_array, time_array, epoch_start)
            })

            # Increase the epoch value
            epoch_start += interval

        # Turn that resample into a dataframe
        return DataFrame(resampled_array, columns=['time', 'value'])

    def get_quantstats_metrics(self):
        try:
            import quantstats as qs
            try:
                returns = self.get_returns()['value']
                returns.index = to_datetime(returns.index, origin=Timestamp(self.start_time, unit='s'), unit='D')
                return qs.reports.metrics(returns, display=False)
            except ValueError as e:
                info_print(e)
        except ImportError:
            raise ImportError(
                "Quantstats not installed. Run 'pip install quantstats' to calculate metrics using Quantstats.")

    def __str__(self):
        return_string = "\n"
        return_string += "Historical Dataframes: \n"

        return_string += "Account History: \n"
        return_string += self.history_and_returns['history'].__str__()
        return_string += "\n"

        return_string += "Account Returns: \n"
        return_string += self.history_and_returns['returns'].__str__()
        return_string += "\n"

        return_string += "Resampled Account Value: \n"
        return_string += self.history_and_returns['resampled_account_value'].__str__()
        return_string += "\n"

        return_string += "Blankly Metrics: \n"
        for i in self.metrics.keys():
            display_name = self.metrics[i]['display_name']
            value = self.metrics[i]['value']
            spaces_needed = 33 - len(display_name)
            user_metrics_line = display_name + ": " + (' ' * spaces_needed) + str(value)
            if display_name[-3:] == "(%)":
                user_metrics_line += "%"
            user_metrics_line += "\n"
            return_string += user_metrics_line

        if self.user_callbacks != {}:
            return_string += "\n"
            return_string += "User Callbacks: \n"
            for i in self.user_callbacks.keys():
                return_string += i + ": " + str(self.user_callbacks[i]) + "\n"

        return return_string



"""
    Abstraction for creating interval driven user strategies
    Copyright (C) 2021  Emerson Dove, Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import datetime
import threading
import time
import typing
import warnings

import pandas as pd

import blankly
from blankly.exchanges.abc_base_exchange import ABCBaseExchange
from blankly.exchanges.exchange import Exchange
from blankly.exchanges.interfaces.abc_base_exchange_interface import ABCBaseExchangeInterface
from blankly.exchanges.interfaces.paper_trade.backtest_controller import BackTestController
from blankly.exchanges.abc_exchange import ABCExchange
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.frameworks.strategy.strategy_state import StrategyState
from blankly.utils.time_builder import time_interval_to_seconds
from blankly.utils.utils import AttributeDict, info_print
from blankly.utils.utils import get_ohlcv_from_list
from blankly.exchanges.strategy_logger import StrategyLogger


# TODO this entire class needs to be fixed it's all fucked
class StrategyBase:
    _exchange: ABCBaseExchange
    interface: ABCBaseExchangeInterface

    def __init__(self, exchange, interface):
        """
        Create a new strategy object. A strategy can be used to run your code live while be backtestable and modular
         across exchanges.
         Args:
             exchange: An exchange object. This can be created by doing something similar to exchange = blankly.Alpaca()

        Function Signatures:
        init(symbol: str, state: blankly.StrategyState)
        price_event(price: float, symbol: str, state: blankly.StrategyState)
        orderbook_event(tick: dict, symbol: str, state: blankly.StrategyState)
        bar_event(bar: dict, symbol: str, state: blankly.StrategyState)
        teardown(blankly.StrategyState)
        """
        self._remote_backtesting = blankly._backtesting
        self._exchange = exchange
        self.interface = interface

        self.ticker_manager = blankly.TickerManager(self._exchange.get_type(), '')
        self.orderbook_manager = blankly.OrderbookManager(self._exchange.get_type(), '')

        self._scheduling_pair = []  # Object to hold a currency and the resolution it's pulled at: ["BTC-USD", 60]

        # Create a cache for the current interface
        self._interface_cache = self.interface
        self._schedulers = []
        self.__variables = {}
        self._hashes = []
        self._orderbook_websockets = []
        self._ticker_websockets = []

        # Initialize backtesting attributes. This only used for sending times to the Strategy/StrategyState
        # This is done because we switch the interface to a paper trade interface
        self.backtesting_controller = None

        # This will throw a warning if they're trying to use an orderbook in the backtest
        self._using_orderbook = False

        # Attempt to report the strategy
        blankly.reporter.export_strategy(self)

        # Create a lock for the teardown so nothing happens while its going on
        self.lock = threading.Lock()

        # This will be updated when the teardown() function completes
        self.torndown = False

    @property
    def variables(self):
        return self.__variables

    def modify_variable(self, callable_: typing.Callable, key, value):
        hashed = hash(callable_)
        self.__variables[hashed][key] = value

    # # TODO these could have some parameters assigned by a super class
    # def add_arbitrage_event(self, callback: typing.Callable, symbols: list, resolution: typing.Union[str, float],
    #                         init: typing.Callable = None, teardown: typing.Callable = None, synced: bool = False,
    #                         variables: dict = None):
    #     """
    #     Add Arbitrage Event - This allows periodic events where prices are gathered asynchronously. When run live at a
    #      small interval, this allows minimal latency when gathering the price of the requested symbols.
    #     """

    def add_price_event(self, callback: typing.Callable, symbol: str, resolution: typing.Union[str, float],
                        init: typing.Callable = None, teardown: typing.Callable = None, synced: bool = False,
                        variables: dict = None):
        """
        Add Price Event. This will provide you with an updated price every time the callback is run
        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the price event for
            resolution: The resolution that the callback will be run - in seconds
            init: Callback function to allow a setup for the strategy variable. Example usages include
                downloading price data before usage
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful
            synced: Sync the function to
            variables: A dictionary to initialize the state's internal values
        """
        self.__custom_price_event(callback=callback, symbol=symbol, resolution=resolution, init=init, synced=synced,
                                  teardown=teardown, variables=variables, type_='price_event')

    def add_bar_event(self, callback: typing.Callable, symbol: str, resolution: typing.Union[str, float],
                      init: typing.Callable = None, teardown: typing.Callable = None, variables: dict = None):
        """
        The bar event sends a dictionary of {open, high, low, close, volume} which has occurred in the interval.
        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the price event for
            resolution: The resolution that the callback will be run - in seconds
            init: Callback function to allow a setup for the strategy variable. Example usages include
                downloading price data before usage
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful
            variables: A dictionary to initialize the state's internal values
        """
        self.__custom_price_event(type_='bar', synced=True, bar=True, callback=callback, symbol=symbol,
                                  resolution=resolution, init=init, teardown=teardown, variables=variables)

    def __custom_price_event(self,
                             type_: str,
                             callback: typing.Callable = None,
                             symbol: str = None,
                             resolution: typing.Union[str, float] = None,
                             init: typing.Callable = None,
                             synced: bool = False,
                             bar: bool = False,
                             teardown: typing.Callable = None, variables: dict = None):
        """
        Add Price Event
        Args:
            type_: The type of event that the price event refers to
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the price event for
            resolution: The resolution that the callback will be run - in seconds
            init: Callback function to allow a setup for the strategy variable. Example usages include
                downloading price data before usage
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful
            synced: Sync the function to
            bar: Get the OHLCV data for a valid exchange interval
            variables: Initial dictionary to write into the state variable
        """
        # Make sure variables is always an empty dictionary if None
        if variables is None:
            variables = {}
        resolution = time_interval_to_seconds(resolution)

        self._scheduling_pair.append([symbol, resolution, type_])

        variables_ = AttributeDict(variables)
        state = StrategyState(self, variables_, symbol, resolution=resolution)

        if resolution < 60:
            # since it's less than 10 sec, we will just use the websocket feed - exchanges don't like fast calls
            self.ticker_manager.create_ticker(self.__idle_event, override_symbol=symbol)
            self._schedulers.append(
                blankly.Scheduler(self.__price_event_websocket, resolution,
                                  initially_stopped=True,
                                  callback=callback,
                                  resolution=resolution,
                                  variables=variables_,
                                  state_object=state,
                                  synced=synced,
                                  init=init,
                                  teardown=teardown,
                                  ohlc=bar,
                                  symbol=symbol)
            )
            exchange_type = self._exchange.get_type()
            self._ticker_websockets.append([symbol, exchange_type, init, state, teardown])
        else:
            # Use the API
            self._schedulers.append(
                blankly.Scheduler(self.__price_event_rest, resolution,
                                  initially_stopped=True,
                                  callback=callback,
                                  resolution=resolution,
                                  variables=variables_,
                                  state_object=state,
                                  synced=synced,
                                  ohlc=bar,
                                  init=init,
                                  teardown=teardown,
                                  symbol=symbol)
            )

        # Export a new symbol to the backend
        blankly.reporter.export_used_symbol(symbol)

    def __idle_event(self, *args, **kwargs):
        """
        Function to skip & ignore callbacks
        """
        pass

    def __price_event_rest(self, **kwargs):
        callback = kwargs['callback']
        symbol = kwargs['symbol']
        resolution = kwargs['resolution']
        variables = kwargs['variables']
        ohlc = kwargs['ohlc']
        state = kwargs['state_object']  # type: StrategyState

        state.variables = variables
        state.resolution = resolution

        if ohlc:
            ohlcv_time = kwargs['ohlcv_time']
            while True:
                # Sometimes coinbase doesn't download recent data correctly
                try:
                    data = self.interface.history(symbol=symbol, to=1, resolution=resolution).iloc[-1].to_dict()
                    if self.interface.get_exchange_type() == "alpaca":
                        break
                    else:
                        if data['time'] + resolution == ohlcv_time:
                            break
                except IndexError:
                    pass
                time.sleep(.5)
            data['price'] = self.interface.get_price(symbol)
        else:
            data = self.interface.get_price(symbol)

        callback(data, symbol, state)

    def __price_event_websocket(self, **kwargs):
        callback = kwargs['callback']
        symbol = kwargs['symbol']
        resolution = kwargs['resolution']
        variables = kwargs['variables']
        ohlc = kwargs['ohlc']
        state = kwargs['state_object']  # type: StrategyState

        if ohlc:
            close_time = kwargs['ohlcv_time']
            open_time = close_time-resolution
            ticker_feed = list(reversed(self.ticker_manager.get_feed(override_symbol=symbol)))
            #     tick        tick
            #      |    ohlcv close                            ohlcv open
            # 0    |   -20          -40            -60        -80
            # newest, newest - 1, newest - 2
            count = 0
            while ticker_feed[count]['time'] > close_time:
                count += 1

            close_index = count

            # Start at the close index to save some iterations
            count = close_index
            while ticker_feed[count]['time'] < open_time:
                count += 1
            # Subtract 1 to put it back inside the range
            count -= 1
            open_index = count

            # Get the latest price that isn't past the timeframe
            last_price = ticker_feed[close_index:][-1]['price']

            data = get_ohlcv_from_list(list(reversed(ticker_feed[close_index:open_index])), last_price)

        else:
            try:
                data = self.ticker_manager.get_most_recent_tick(override_symbol=symbol)['price']
            except TypeError:
                info_print("No valid data yet - using rest.")
                data = self.interface.get_price(symbol)

        state.variables = variables
        state.resolution = resolution

        callback(data, symbol, state)

    @staticmethod
    def __orderbook_event(tick, symbol, user_callback, state_object):
        user_callback(tick, symbol, state_object)

    def add_orderbook_event(self, callback: typing.Callable, symbol: str, init: typing.Callable = None,
                            teardown: typing.Callable = None, variables: dict = None):
        """
        Add Orderbook Event
        Args:
            callback: The price event callback that will be added to the current ticker and run at the proper resolution
            symbol: Currency pair to create the orderbook for
            init: Callback function to allow a setup for the strategy variable. This
                can be used for accumulating price data
            teardown: A function to run when the strategy is stopped or interrupted. Example usages include liquidating
                positions, writing or cleaning up data or anything else useful:
            variables: A dictionary to initialize the state's internal values
        """
        # Make sure variables is always an empty dictionary if None
        if variables is None:
            variables = {}
        self._scheduling_pair.append([symbol, None])
        callback_hash = hash((callback, symbol))
        if callback_hash in self._hashes:
            raise ValueError("A callback of the same type and resolution has already been made for the ticker: "
                             "{}".format(symbol))
        else:
            self._hashes.append(callback_hash)
        self.__variables[callback_hash] = AttributeDict(variables)
        state = StrategyState(self, self.__variables[callback_hash], symbol=symbol)

        variables = self.__variables[callback_hash]

        # since it's less than 10 sec, we will just use the websocket feed - exchanges don't like fast calls
        self.orderbook_manager.create_orderbook(self.__orderbook_event, initially_stopped=True,
                                                override_symbol=symbol,
                                                symbol=symbol,
                                                user_callback=callback,
                                                variables=variables,
                                                state_object=state)

        exchange_type = self._exchange.get_type()
        self._orderbook_websockets.append([symbol, exchange_type, init, state, teardown])

        # Set this to true so that we can throw a warning in the backtest
        self._using_orderbook = True

    def start(self):
        """
        Run your model live!

        Simply call this function to take your strategy configuration live on your exchange
        """
        if self._remote_backtesting:
            warnings.warn("Aborted attempt to start a live strategy a backtest configuration")
            return
        for i in self._schedulers:
            kwargs = i.get_kwargs()
            if kwargs['init'] is not None:
                kwargs['init'](kwargs['symbol'], kwargs['state_object'])
            i.start()

        for i in self._orderbook_websockets:
            # Index 2 contains the initialization function for the assigned websockets array
            if i[2] is not None:
                i[2](i[0], i[3])
            self.orderbook_manager.restart_ticker(i[0], i[1])

        for i in self._ticker_websockets:
            # The initialization function should have already been called for ticker websockets
            # Notice this is different from orderbook websockets because these are put into the scheduler
            self.ticker_manager.restart_ticker(i[0], i[1])

    def teardown(self):
        self.lock.acquire()
        for i in self._schedulers:
            i.stop_scheduler()
            kwargs = i.get_kwargs()
            teardown = kwargs['teardown']
            state_object = kwargs['state_object']
            if callable(teardown):
                teardown(state_object)

        for i in self._orderbook_websockets:
            self.orderbook_manager.close_websocket(override_symbol=i[0], override_exchange=i[1])
            # Call the stored teardown
            teardown_func = i[4]
            if callable(teardown_func):
                teardown_func(i[3])

        for i in self._ticker_websockets:
            self.ticker_manager.close_websocket(override_symbol=i[0], override_exchange=i[1])
        self.lock.release()

        # Show that all teardowns have finished
        self.torndown = True

    def time(self) -> float:
        raise NotImplementedError

    def backtest(self,
                 to: str = None,
                 initial_values: dict = None,
                 start_date: typing.Union[str, float, int] = None,
                 end_date: typing.Union[str, float, int] = None,
                 save: bool = False,
                 settings_path: str = None,
                 callbacks: list = None,
                 **kwargs
                 ) -> BacktestResult:
        raise NotImplementedError




from blankly.exchanges.exchange import Exchange
from blankly.exchanges.interfaces.oanda.oanda_api import OandaAPI
from blankly.exchanges.auth.auth_constructor import AuthConstructor


class Oanda(Exchange):
    def __init__(self, portfolio_name=None, keys_path="keys.json", settings_path=None):
        Exchange.__init__(self, "oanda", portfolio_name, settings_path)

        # Load the auth from the keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'oanda', ['PERSONAL_ACCESS_TOKEN', 'ACCOUNT_ID', 'sandbox'])

        sandbox = super().evaluate_sandbox(auth)

        keys = auth.keys
        calls = OandaAPI(personal_access_token=keys['PERSONAL_ACCESS_TOKEN'],
                         account_id=keys['ACCOUNT_ID'],
                         sandbox=sandbox)

        # Always finish the method with this function
        super().construct_interface_and_cache(calls)

    def get_exchange_state(self):
        return self.interface.get_fees()

    def get_asset_state(self, symbol):
        return self.interface.get_account(symbol)

    def get_direct_calls(self) -> OandaAPI:
        return self.calls



from setuptools import find_packages, setup
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, './README.md'), encoding='utf-8') as f:
    long_description = f.read()

"""
Build Info:
python3 -m build
twine upload dist/*
"""

setup(
    name='blankly',  # How you named your package folder (MyLib)
    packages=find_packages(),
    # packages=['blankly'],  # Potentially should be the same thing as name
    version='v1.16.17-beta',
    license='lgpl-3.0',  # Licenses: https://help.github.com/articles/licensing-a-repository
    description='Rapidly build & deploy cross-exchange trading bots',  # Give a short description about your library
    long_description=long_description,
    long_description_content_type='text/markdown',
    ext_modules=[
        # Extension("blankly.deployment.hello", [path.join("blankly", "deployment", "hello.cpp")])  # Disabled extension
    ],
    author='blankly',
    entry_points={'console_scripts': ['blankly = blankly.deployment.cli:main']},
    author_email='contact@blankly.finance',
    url='https://github.com/Blankly-Finance/Blankly',  # Could be github or website
    # download_url='https://github.com/EmersonDove/Blankly/archive/v0.1.1-alpha.tar.gz',
    keywords=['Crypto', 'Stocks', 'Quantitative Finance', 'Exchanges', 'Bot'],  # Keywords
    install_requires=[
        'alpaca-trade-api >= 1.4.2',
        'bokeh >= 2.4.2',
        'dateparser >= 1.1.0',
        'newtulipy >= 0.4.6',
        'numpy >= 1.21.4',
        'pandas >= 1.1.5',
        'pandas_market_calendars >= 3.2',
        'python-binance >= 1.0.15',
        'requests >= 2.26.0',
        'websocket-client >= 1.2.1',
    ],
    classifiers=[
        # Possible: "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
)



"""
    Interface for ticker objects.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc


class ABCExchangeWebsocket(abc.ABC):
    """
    Each of these functions are required to be implemented for interaction with the ticker manager class.
    """
    @abc.abstractmethod
    def is_websocket_open(self):
        pass

    @abc.abstractmethod
    def append_callback(self, obj):
        pass

    @abc.abstractmethod
    def get_most_recent_tick(self):
        pass

    @abc.abstractmethod
    def get_most_recent_time(self):
        pass

    @abc.abstractmethod
    def get_time_feed(self):
        pass

    @abc.abstractmethod
    def get_feed(self):
        pass

    @abc.abstractmethod
    def get_response(self):
        pass

    @abc.abstractmethod
    def close_websocket(self):
        pass

    @abc.abstractmethod
    def restart_ticker(self):
        pass



"""
    This file shows how multicore bots can be spawned & used to carry out intensive but separate tasks
"""

import time
import blankly
# Have feedback? Use this: https://forms.gle/4oAjG9MKRTYKX2hP9


class Bot(blankly.BlanklyBot):
    def main(self, args):
        """
        Main function to write any general analysis or bot management logic
        """
        # Create a demo ticker object. The price event function will run at every tick
        self.ticker_manager.create_ticker(callback=self.price_event)

        # Add a heartbeat example to report to GUI or main
        self.update_state("heartbeat", 0)

        # This will work on any supported exchange
        print("Interface call: " + str(self.interface.get_fees()))

        # You can also bypass the interface and make calls directly to the exchange:
        try:
            # Because this is direct, some exchanges have different commands, so this demo is wrapped in try/except
            print("Direct call: " + str(self.direct_calls.get_fees()))
        except AttributeError:
            # If it doesn't have that function we just complain about it and move on
            print("Chosen exchange doesn't have that function")

        while True:
            # This demonstrates a way to change the state. This is just reporting to the main.
            current_heartbeat = self.get_state()["heartbeat"]
            self.update_state("heartbeat", current_heartbeat + 1)
            time.sleep(1)

    def price_event(self, tick):
        """
        This is called on price update events from the ticker feeds
        """
        # Example of updating the price state for the GUI
        self.update_state("Price", str(tick["price"]))
        # Show these new ticks to the console
        print("New price tick at: " + str(tick["price"]))


if __name__ == "__main__":
    """
    Easily setup and run a model on any supported exchange
    """

    # This creates an authenticated exchange and chooses the first portfolio in the keys.json file
    # You can use portfolio_name="my cool portfolio" if want a certain one
    portfolio = blankly.CoinbasePro()  # You could also use blankly.Binance()

    # Create the bot and add it to run as a coinbase_pro bitcoin model.
    bot = Bot()
    portfolio.append_model(model=bot, symbol="BTC-USD", args=[])

    # This starts the main() function of the model and puts it on a different process (computer core)
    # The main() function is in the class above
    portfolio.start_models(symbol="BTC-USD")

    # Now other processes can be created or just continue with this one.
    while True:
        # Print the state from the model we just started every second
        state = portfolio.get_full_state("BTC-USD")
        blankly.utils.pretty_print_json(state)
        time.sleep(1)
